﻿

using System;
using System.Drawing;
using System.Windows.Forms;
using System.Collections.Generic;
using System.Text.RegularExpressions;


namespace MyFx
{
	/// <summary>
	/// Description of EffectParams.
	/// </summary>
	public partial class EffectParams : Form
	{
		public EffectParams()
		{
			//
			// The InitializeComponent() call is required for Windows Forms designer support.
			//
			InitializeComponent();
			
			//
			// TODO: Add constructor code after the InitializeComponent() call.
			//
		}
#region "Properties"
		private string m_propname;
		public string PropName
		{
			get { return this.m_propname; }
			set { this.m_propname = value; this.Text/*.Insert(0, value + " ")*/ = value + " " + this.Text; }
		}

		private Dictionary<string, byte> m_effectChoices;
		public Dictionary<string, byte> effectChoices
		{
			get { return this.m_effectChoices; }
			set
			{
				this.m_effectChoices = value;
				this.effectComboBox.Items.Clear();
				foreach (string key in value.Keys) this.effectComboBox.Items.Add(key);
			}
		}
//		private bool WantDebug = false;
//		private string logfile = string.Empty;
//		private bool first_debug = true;
//		private void LogMsg(string msg)
//		{
////			if (!this.WantDebug) return; // caller should check this
//			if (string.IsNullOrEmpty(logfile)) logfile = System.IO.Path.Combine(System.IO.Path.GetDirectoryName(System.Reflection.Assembly.GetExecutingAssembly().Location), "..\\..\\zdebug.log");
//			if (this.first_debug) MessageBox.Show("debug output is going to '" + logfile + "'");
//			this.first_debug = false;
//			System.IO.File.AppendAllText(logfile, msg + "\r\n"); //CAUTION: not thread-safe; must be serialized
//		}
		public byte effectBox
		{
			get
			{
				//LogMsg("set effect box, sel inx " + this.effectComboBox.SelectedIndex);
				if (this.effectComboBox.SelectedIndex != -1)
				{
					//LogMsg("sel text " + this.effectComboBox.Items[this.effectComboBox.SelectedIndex]);
					//LogMsg("assoc val " + this.m_effectChoices[(string)this.effectComboBox.Items[this.effectComboBox.SelectedIndex]]);
				}
				if (this.effectComboBox.SelectedIndex != -1) return this.m_effectChoices[(string)this.effectComboBox.Items[this.effectComboBox.SelectedIndex]];
				return 0;
			}
			set
			{
				int selinx = -1; //NOTE: don't set .SelectedIndex directly (combo box complains if goes past end of list)
				foreach (byte val in this.m_effectChoices.Values)
				{
					++selinx;
					if (val == value) { this.effectComboBox.SelectedIndex = selinx; return; }
				}
				this.effectComboBox.SelectedIndex = -1;
			}
		}

		private Dictionary<string, byte> m_textChoices;
		public Dictionary<string, byte> textChoices
		{
			get { return this.m_textChoices; }
			set
			{
				this.m_textChoices = value;
				this.textComboBox.Items.Clear();
				foreach (string key in value.Keys) this.textComboBox.Items.Add(key);
			}
		}
		public byte textBox
		{
			get
			{
				if (this.textComboBox.SelectedIndex != -1) return this.m_textChoices[(string)this.textComboBox.Items[this.textComboBox.SelectedIndex]];
				return 0;
			}
			set
			{
//				for (int i=0; i<this.m_textChoices.Items.Count; ++i)
//					if (this.m_textChoices.Values[i] == value)
//					{
//						this.m_textChoices.SelectedIndex = i;
//						return;
//					}
				int selinx = -1; //NOTE: don't set .SelectedIndex directly (combo box complains if goes past end of list)
				foreach (byte val in this.m_textChoices.Values)
				{
					++selinx;
					if (val == value) { this.textComboBox.SelectedIndex = selinx; return; }
				}
				this.textComboBox.SelectedIndex = -1;
			}
		}

		private byte m_paramBox;
		public byte paramBox
		{
			get { return this.m_paramBox; }
			set
			{
				this.m_paramBox = value;
				this.paramTextBox.Text = value.ToString();
			}
		}

		private Color m_colorBox;
		public Color colorBox
		{
			get { return this.m_colorBox; }
			set
			{
				this.m_colorBox = value;
				if ((value.A == 0) && (value.R == 0) && (value.G == 0) && (value.B == 0)) this.m_colorBox = Color.FromArgb(255, 0, 0, 0); //kludge: init to opaque
				this.colorTextBox.Text = this.m_colorBox.A.ToString() + ", " + this.m_colorBox.R.ToString() + ", " + this.m_colorBox.G.ToString() + ", " + this.m_colorBox.B.ToString();
			}
		}
		private static Regex colorparse = new Regex("^\\s*(\\d+),\\s*(\\d+),\\s*(\\d+),\\s*(\\d+)$"); //A, R, G, B
#endregion

#region "Event handlers"
		private void ColorBrowse(object sender, EventArgs e)
		{
			ColorChooser.ColorChooser.ColorChooser3 cc = new ColorChooser.ColorChooser.ColorChooser3();
			cc.Text = "Select " + this.m_propname + " Color";
			cc.Color = this.colorBox;
			if (cc.ShowDialog(this) == System.Windows.Forms.DialogResult.OK) this.colorBox = cc.Color;
			cc.Dispose();
		}

		private void FormClose(object sender, FormClosingEventArgs e)
		{
			if (this.dirty && (((Control)sender).Name == "cancelButton"))
				switch (MessageBox.Show("Save changes?", this.Text, MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question))
				{
					case DialogResult.No: return;
					case DialogResult.Yes: this.DialogResult = DialogResult.OK; return;
					case DialogResult.Cancel: e.Cancel = true; return;
				}
		}

		private bool dirty = false;
		private void DataChanged(object sender, EventArgs e)
		{
			if (((Control)sender).Name == "colorTextBox")
			{
				System.Text.RegularExpressions.Match parts = colorparse.Match(((TextBox)sender).Text);
				if (parts.Success && (parts.Groups.Count == 4+1))
				{
					int a = int.Parse(parts.Groups[1].Value);
					int r = int.Parse(parts.Groups[2].Value);
					int g = int.Parse(parts.Groups[3].Value);
					int b = int.Parse(parts.Groups[4].Value);
					if ((a < 256) && (r < 256) && (g < 256) && (b < 256))
					{
						this.colorBox = Color.FromArgb(a, r, g, b);
						this.dirty = true;
						return;
					}
				}
				MessageBox.Show("Invalid color.  Format is #, #, #, # (each # is 0..255)", this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
				return;
			}
			if (((Control)sender).Name == "paramTextBox")
			{
				byte val;
				if (byte.TryParse(((TextBox)sender).Text, out val))
				{
					this.paramBox = val;
					this.dirty = true;
					return;
				}
				MessageBox.Show("Invalid other parameter.  Must be number 0..255", this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
				return;
			}
			this.dirty = true;
		}
#endregion

	}
}
