'
'Vixen Effects Generator
'
'Copyright (c) 2011-2012 Don Julien.  djulien@thejuliens.net
'These effects may be freely distributed and used for non-commerical purposes only.
' See the license agreement terms below.
'
'Revision history:
'0  6/31/11  DJ  first version; it's time to do some RGB grid animation in Vixen!
'1.0  7/?/11  DJ  first stable/usable version
'1.1 2/21/12  DJ  cleaned up for general release; move hard-coded stuff to custom fx (child class)
'1.2 9/21/12  DJ  a little more clean up for interim release
'
' SOFTWARE LICENSE AGREEMENT:
' TODO: GPL?
' This extended software is licensed under the Creative Commons Attribution-Noncommercial-Share Alike
' 3.0 Unported License.  To view a copy of this license, visit http://creativecommons.org/licenses/by-nc-sa/3.0/
' or send a letter to Creative Commons, 171 Second Street, Suite 300, San Francisco, California, 94105, USA.
'
' Basically, you may use this software freely for your own personal, non-commerical use, as long as you also share
' freely any changes, adaptations, or improvements you make.  Of course, the standard conditions also apply:
' THIS SOFTWARE IS PROVIDED "AS IS".  NO EXPRESS, IMPLIED, STATUTORY OR OTHER WARRANTIES
' OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE APPLY TO THIS SOFTWARE.  AUTHOR
' SHALL NOT BE LIABLE FOR SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES, FOR ANY REASON OR
' UNDER ANY CIRCUMSTANCES WHATSOEVER.  THIS SOFTWARE MUST NOT BE USED FOR CRITICAL APPLICATIONS
' SUCH AS HEAVY MACHINERY, LIFE SUPPORT, ETC.  YOU USE THIS SOFTWARE AT YOUR OWN RISK.
' But if you have problems or questions, let me know and I'll try to help if I can.

'Usage:
'Place this compiled .DLL into the Vixen/Plugins/Output folder.
'This output plug-in will execute the Effects script to generate new channel values,
'then pass those to the target output plug-in(s) as if Vixen had done so.
'
'Note about AddHandler vs. WithEvents/Handles:
'Even though AddHandler may be more efficient, I prefer WithEvents/Handles for several reasons:
'- it allows the logic to point to the control, instead of the control pointing to the logic
'- the Form Designer may delete the control and, hence, the link to the logic; if the link is with the logic, this won't happen
'- AddHandler requires manually editing code that was generated automatically by the Form Designer, which seems to be a violation of the underlying principle
'
'OPERATION:
'- Setup checks for timing track (sequential frame#s); if not found, it is added to sequence
'- during playback, timing track forces Vixen to invoke this output plug-in for each frame
'- for each target plug-in, channel values are read from sequence, Effects applied, and new values passed to output plug-in
'- changing the execution point will loose sync between sequence and audio; there are a couple of ways to deal with this:
'  use a hacked up/longer audio track, try to manipulate Vixen execution point, or control audio playback from this plug-in instead of Vixen

'what this does:
'1. sits as layer between Vixen and output plug-ins
'2. reads channel values from Sequence, wraps as GDI bitmap and calls plug-ins
'3. writes updated GDI bitmap back to Sequence
'modes: read on/off, write on/off
'could be used to generate brightness normalization curves (see budude thread)
'TBD which parts (if any) are redundant with Vixen 3.x?
'probably will add a little code to make it compatible next year
'desired paradigm: like openSCAD (precise scripting rather than tedious mousing)
'this code can run 15k channels (maybe more) on a Win XP VM with < 512 MB RAM no problem
'script warnings cause error; use script editor to compile/check before hand

'TODO:
'fix pkugin name
'fix plugin vis from bkg thread
'fill in missing vix evts
'needs a lot of source code clean-up
'mem leak?
'-write effects script file to Vixen sequence/profile?
'-implement MP3 control
'-add more effects
'-re-wrap as IDE plug-in
 
'The overall VB project/solution structure is based on an earlier Vixen 2.1/2.5-compatible design pattern.
'The effects generation itself is fairly obvious based on how channels work in Vixen.
'
'for some sketchy info about Vixen plugins, see:
'http://www.doityourselfchristmas.com/forums/showthread.php?t=3929
'http://www.doityourselfchristmas.com/forums/showthread.php?t=50324
'http://www.doityourselfchristmas.com/forums/showthread.php?t=722

Option Strict On
Option Explicit On
Option Compare Text

Imports System
'Imports System.Collections.Generic
Imports System.IO.File
Imports System.Text
Imports System.Text.RegularExpressions
'Imports System.Runtime.InteropServices
Imports System.Windows.Forms
Imports System.Threading
Imports System.Xml
Imports System.Reflection
Imports System.Drawing


Imports Vixen 'this will be 2.1 or 2.5, depending on the project references

'dummy interface to force Vixen to load this DLL:
'Public Interface IOutputPlugIn
'	Sub DummyFunction()
'End Interface

Namespace FxGen


#Region "Debug wrapper"
#If False
' Vixen doesn't report exceptions within plug-in ctor, so this wrapper allows that to happen
Public Class FxGenDiagnosticsWrapper
	Implements Vixen.IEventDrivenOutputPlugIn 'Vixen events

	Private m_fxgen As VixFxGenPlugin
	Sub New() 'this one is needed for Vixen to be able to instantiate the plug-in
'		MyBase.New'(Nothing)
		Try
			Me.m_fxgen = New VixFxGenPlugin()
		Catch exc As Exception
			MessageBox.Show(String.Format("plugin ctor exc: {0}", exc.ToString), "ERROR", System.Windows.Forms.MessageBoxButtons.OKCancel, System.Windows.Forms.MessageBoxIcon.Exclamation)
		End Try
	End Sub

	Private ReadOnly Property Name_any As String Implements Vixen.IEventDrivenOutputPlugIn.Name
		Get
			If Me.m_fxgen IsNot Nothing Then Return DirectCast(Me.m_fxgen, Vixen.IEventDrivenOutputPlugIn).Name
			Return "FxGen Wrapper"
		End Get
	End Property

	Private ReadOnly Property Description_any As String Implements Vixen.IEventDrivenOutputPlugIn.Description
		Get
			If Me.m_fxgen IsNot Nothing Then Return DirectCast(Me.m_fxgen, Vixen.IEventDrivenOutputPlugIn).Description
			Return "FxGen diagnostics wrapper"
		End Get
	End Property

	Private ReadOnly Property Author_any As String Implements Vixen.IEventDrivenOutputPlugIn.Author
		Get
			If Me.m_fxgen IsNot Nothing Then Return DirectCast(Me.m_fxgen, Vixen.IEventDrivenOutputPlugIn).Author
			Return "DIYC/djulien"
		End Get
	End Property

	Private ReadOnly Property HardwareMap_any As Vixen.HardwareMap() Implements Vixen.IEventDrivenOutputPlugIn.HardwareMap
		Get
			If Me.m_fxgen IsNot Nothing Then Return DirectCast(Me.m_fxgen, Vixen.IEventDrivenOutputPlugIn).HardwareMap
			Return New HardwareMap(0  - 1){}
		End Get
	End Property

#If VIXEN = 21
'interface broken (return type different between 2.1 and 2.5)
	Function StartUp_21() As List(Of system.Windows.Forms.Form) Implements Vixen.IEventDrivenOutputPlugIn.StartUp
		If Me.m_fxgen IsNot Nothing Then Return DirectCast(Me.m_fxgen, Vixen.IEventDrivenOutputPlugIn).Startup
		Return Nothing
	End Function

'interface strange (this one needs to be implemented 2x)
	Overrides Function ToString() As String Implements Vixen.IOutputPlugIn.ToString
		Return Me.Name_any 'DirectCast(Me.m_fxgen, Vixen.IEventDrivenOutputPlugIn).Name
	End Function
#End If

#If VIXEN = 25
'interface broken (return type different between 2.1 and 2.5)
	Sub StartUp_25() Implements Vixen.IEventDrivenOutputPlugIn.StartUp
		If Me.m_fxgen IsNot Nothing Then DirectCast(Me.m_fxgen, Vixen.IEventDrivenOutputPlugIn).Startup()
	End Sub
#End If

	Private Function ToString_any() As String Implements Vixen.IEventDrivenOutputPlugin.ToString', Vixen.IOutputPlugIn.ToString
		Return Me.Name_any 'DirectCast(Me.m_fxgen, Vixen.IEventDrivenOutputPlugIn).Name
	End Function

	Private Sub Initialize_any(seq As Vixen.IExecutable, setupData As Vixen.SetupData, setupNode As XmlNode) Implements Vixen.IEventDrivenOutputPlugIn.Initialize
		If Me.m_fxgen IsNot Nothing Then DirectCast(Me.m_fxgen, Vixen.IEventDrivenOutputPlugIn).Initialize(seq, setupData, setupNode)
	End Sub

	Private Sub Setup_any() Implements Vixen.IEventDrivenOutputPlugIn.Setup
		If Me.m_fxgen IsNot Nothing Then DirectCast(Me.m_fxgen, Vixen.IEventDrivenOutputPlugIn).Setup()
	End Sub

	Private Sub Event_any(channelValues As Byte()) Implements Vixen.IEventDrivenOutputPlugIn.[Event]
		If Me.m_fxgen IsNot Nothing Then DirectCast(Me.m_fxgen, Vixen.IEventDrivenOutputPlugIn).[Event](channelValues)
	End Sub

	Private Sub Shutdown_any() Implements Vixen.IEventDrivenOutputPlugIn.Shutdown
		If Me.m_fxgen IsNot Nothing Then DirectCast(Me.m_fxgen, Vixen.IEventDrivenOutputPlugIn).Shutdown()
	End Sub
End Class
#End If
#End Region


'main implementation class:
'NOTE: this is an abstract base class; custom effects must be added to a derived class before it can be used
Public MustInherit Partial Class FxGenPluginBase
'?	Inherits PluginWrapper 'FxGen.FxBase 'FxGen external effects linkage
	Implements Vixen.IEventDrivenOutputPlugIn 'Vixen plug-in API

#Region "Internal data"
'config info from Vixen:
	Private m_setupNode As XmlNode = Nothing
	Public VixenInfo As New FxGen.VixenInfo() 'setup data, sequence info from Vixen
'	Private m_plugins As New FxGen.SortedList_OfPlugInfo() '(Of String, PlugInfo)

'custom config info:
'	Private m_logger As New FxGen.FxLogger() 'debug/performance log
'	private shared Logge as New Logger()
'	Private m_fxfile As New EffectsFileInfo(Me.m_logger)
	Private m_virtseq As Vixen.IExecutable = Nothing '.EventSequence = Nothing 'inner (virtual) Vixen sequence/profile
	Private m_rdseq As Boolean = True 'whether to use existing channel values
	Private m_wrseq As Boolean = True 'whether to save generated channel values
	Private m_startch As Integer = 0 'start channel when copying to virt to physical channels
	Private m_audio As Boolean = False 'whether to take over audio control
	Private ReadOnly m_bkgproc As Boolean = False 'True 'run processing as background thread; helps keep Vixen responsive
'	Private m_pixtype As String = String.Empty 'type of pixels in canvas
'	Private m_chorder As String = String.Empty 'order of channels in canvas
'	Private m_gridsize As New Size(0, 0) 'canvas geometry
'	Private m_frbuf As Byte() = Nothing
'	Private m_chmapper As channelmapper = Nothing
'	Public ReadOnly ChannelMapper As ChannelMapperBase = Nothing

'run-time state:
	Private Enum WkerState
		Init 'initalized
		Startup 'ready/waiting for startup message
		NextFrame 'ready/waiting for next frame event
		Shutdown 'ready/waiting for shutdown message
		AllDone 'shutdown completed successfully
		ErrorOccurred
	End Enum
	Private m_cfgok As Boolean = False 'remember whether config is okay
	Private m_wkerstate As WkerState = WkerState.AllDone 'status of worker thread
	Private m_wkerthread As Thread = Nothing 'background processing thread
	Private m_wkersync As New AutoResetEvent(False) 'thread sync

'#If False 'need to comment this out for initial compile
'	Private m_fx As New MyFx.CustomFx()
'#End If
'	Private m_shuttingdown As Boolean = True
	Private m_fxDataChanged As Boolean = False
'	Private m_plugins As New ListOf_PluginWrapper() '(Of String, PluginWrapper)
	Private m_rendered As Integer 'remember which frame was rendered
	Public FrameBuf As Byte() = Nothing 'channel values for next Vixen frame (1 ahead)
	Private m_dialogList As New List(Of Form)' = List(Of Form)
'	Private m_TagsByFrame As New Dictionary(Of Integer, List(Of FxGen.FrameTag)) 'frame tags/notes/cues, indexed by start/stop frame
'	Private m_TagsByValue As New Dictionary(Of Integer, List(Of FxGen.FrameTag)) 'frame tags/notes/cues, indexed by start/stop frame
'	Private m_acttags As New List(Of String) 'list of active tags/notes/cues
#End Region

#Region "Vixen version-dependent stuff"
'KLUDGE: Vixen 2.5.0.8 is incompatible with 2.1.1.0; IOutputPlugin.Startup() changed return types, and ToString() is a little strange also.
'Since the published interface breaks compatibility going from 2.1 to 2.5, an ugly work is needed around to fix the problem.

#If VIXEN = 21
	Private Const m_vixver As String = "2.1.x"
'interface broken (return type different between 2.1 and 2.5)
	Function StartUp_21 As List(Of system.Windows.Forms.form) Implements Vixen.IEventDrivenOutputPlugIn.StartUp
		Me.StartUp_any
		Return Me.m_dialogList
	End Function

'interface strange (this one needs to be implemented 2x)
	Public Overrides Function ToString() As String Implements Vixen.IOutputPlugIn.ToString
		Return Me.Name_any
	End Function
#End If

#If VIXEN = 25
	Private Const m_vixver As String = "2.5.x"
'interface broken (return type different between 2.1 and 2.5)
	Sub StartUp_25 Implements Vixen.IEventDrivenOutputPlugIn.StartUp
		Me.StartUp_any
	End Sub
#End If
#End Region

#Region "Instantiation"
	Public Sub New()
'		MyBase.New(Nothing) 'grandparent; dummy linkage to satisfy class hierarchy (this is the top-level object)
'TODO: is it better to statically allocate memory in declares, or dynamically declare them here? (it probably doesn't matter)
'		Try 'On Error GoTo errh 'an exception in here will cause Vixen to say it couldn't find the plug-in; add a little error handling to make that less confusing
'!		Me.m_logger = New FxGen.FxLogger
'?		Me.m_logger.StartTime = DateTime.Now 'for elapsed time in log
'!		Me.m_dialogList = New List(Of Form)
'!		Me.VixenInfo = New FxGen.VixenInfo
'!		Me.m_setupNode = Nothing
'!		Me.m_fxfile = New EffectsFileInfo(Me.m_logger) 'String.Empty
'!		Me.m_audio = False
'!		Me.m_plugins = New SortedList_OfPlugInfo
'!		Me.VixenInfo.Channels = New List(Of Channel)
'!		Me.m_acttags = New List(Of String)
'!		Me.m_frtags = New Dictionary(Of Integer, List(Of FxGen.FrameTag))
'!		Me.m_bkgproc = False 'TODO
'!		Me.m_wkerthread = Nothing
'!		Me.m_wkerstate = WkerState.AllDone
'!		Me.m_wkersync = New AutoResetEvent(False)
'!		Me.m_framebuf = Nothing
'no worky:
'		Catch exc As Exception
'			Me.m_logger.ReportError("ctor", exc)
'			MessageBox.Show(String.Format("plugin ctor exc: {0}", exc.ToString), "ERROR", System.Windows.Forms.MessageBoxButtons.OKCancel, System.Windows.Forms.MessageBoxIcon.Exclamation)
'		End Try
	End Sub

'	Public Sub Dispose(ByVal disposing As Boolean)
'		If disposing = False Then Return
'		If Me.m_plugin IsNot Nothing Then Me.m_plugin.Shutdown
''		Me.m_plugin.Dispose
'		Me.m_plugin = Nothing
''		MyBase.Dispose(disposing)
'	End Sub
#End Region

#Region "Vixen Properties"
	Public MustOverride Function MyName() As String 'require override to avoid confusion in Plugins list
	Private ReadOnly Property Name_any As String Implements Vixen.IEventDrivenOutputPlugIn.Name
		Get
			Return Me.MyName().Replace("/", "&") 'String.Format("FxGen for Vixen {0}", m_vixver) 'CAUTION: don't use "/" in name!
		End Get
	End Property

	Public Overridable Function MyDesc() As String
		Return "Generic base for Scripted Effects Generator plug-in"
	End Function
	Private ReadOnly Property Description_any As String Implements Vixen.IEventDrivenOutputPlugIn.Description
		Get
			Return Me.MyDesc()
		End Get
	End Property

	Public MustOverride Function MyAuthor() As String 'require override to encourage attribution
	Private ReadOnly Property Author_any As String Implements Vixen.IEventDrivenOutputPlugIn.Author
		Get
			Return Me.MyAuthor() '"DIYC.djulien"
		End Get
	End Property

'from Vixen tutoral:
'"The plug-in is expected to provide an array of which PortType resources it is using by index. Currently
'only supports four serial And three parallel ports."
	Private ReadOnly Property HardwareMap_any As Vixen.HardwareMap() Implements Vixen.IEventDrivenOutputPlugIn.HardwareMap
		Get
			Return New HardwareMap(0 -1) {}
		End Get
	End Property

'from Vixen tutoral:
'"Vixen expects that a plug-in will provide a ToString override returning the Name member. The return value
'of ToString will be what's displayed in any listboxes in which the plug-In Is listed."
	Private Function ToString_any() As String Implements Vixen.IEventDrivenOutputPlugin.ToString', Vixen.IOutputPlugIn.ToString
		Return Me.Name_any
	End Function
#End Region

#Region "Vixen event handlers"
'from Vixen tutoral:
'"Called anytime Vixen needs to make sure the plug-in's setup or other initialization is up to date.
'Initialize is called *before* the plug-in is setup, before sequence execution, and other times. It's called
'from multiple places at any time, therefore the plug-in can make no assumptions about the state of the
'program or sequence due to a call to Initialize.
'	channels: An array of Channel objects representing the channels present in the owning sequence.
'	setupData: A SetupData reference that provides some plug-incentric convenience methods. It can be safely ignored.
'	setupNode: An XmlNode representing the root of the plug-in's setup data in the sequence. Please see the Vixen.Xml class for some convenience methods."
	Private Sub Initialize_any(seq As Vixen.IExecutable, setupData As Vixen.SetupData, setupNode As XmlNode) Implements Vixen.IEventDrivenOutputPlugIn.Initialize
'		Dim errmsg As String = String.Empty
		Try
		Logger.StartTime = Date.Now 'basis for elapsed time shown in log
'?		Me.m_logger.StartTime = DateTime.Now 'for elapsed time in log
10:		Dim reinit As Boolean = (Me.m_setupNode IsNot Nothing) 'paranoid: check for re-init
'save Vixen info for later:
'11:		Me.VixenInfo.Exec = seq
		Me.VixenInfo.Seq = seq 'could be sequence or profile 'DirectCast(seq, Vixen.EventSequence)
13:		Me.m_setupNode = setupNode
12:		Me.VixenInfo.SetupData = setupData
'read the logging-related config options ahead of time so they can take effect sooner
21:		Logger.LogFile = Me.VixenInfo.Rel2Abs(Me.VixenInfo.SetupData.GetString(Me.m_setupNode, "LogFile", String.Empty))
22:		Logger.ErrPause = Me.VixenInfo.SetupData.GetBoolean(Me.m_setupNode, "PauseOnError", False)
23:		Logger.WantTrace = Me.VixenInfo.SetupData.GetBoolean(Me.m_setupNode, "DataLogging", False)
'load data that cannot change during plug-in lifespan:
24:		Dim ver As Version = Reflection.Assembly.GetEntryAssembly.GetName.Version 'GetCallingAssembly=mscorlib 'GetExecutingAssmebly=this dll
25:		If Logger.WantTrace Then Logger.LogMsg(String.Format("========== {5} ver 1.0, Vixen {3}.{4} {0}init at {1:MM-dd-yyyy HH:mm:ss}.{2:D3} ==========", IIf(reinit, "re-", String.Empty), DateTime.Now, DateTime.Now.TimeOfDay.Milliseconds, ver.Major, ver.Minor, Me.Name_any())) 'ver.MajorRevision, ver.MinorRevision,
'		Try
27:		Me.LoadChannels()
		Me.LoadConfig(False) 'NOTE: config data can change later (after Setup)
'		Me.m_logger.LogMsg(String.Format("log {0} {1}", Me.m_logger.LogFile, Me.m_logger.WantTrace))
		Catch exc As Exception
			Logger.ReportError("Init", exc)
		End Try
'		If Not String.IsNullOrEmpty(errmsg) Then Throw New Exception(errmsg) 'throw channel-related errors back to Vixen
	End Sub

'from Vixen tutoral:
'"Called when the user has requested to setup the plug-in instance."
'my additional comments:
'Entry in xml file looks something like this:
'<PlugIn name="Renard Dimmer (modified)" key="935429715" id="0" enabled="False" from="1" to="296">
'This event handler reads config params, displays to user, and then writes back to config file if params changed.
'This is the only function that writes config data.
	Public MustOverride Function EnumFuncCodes() As Dictionary(Of String, Integer)
	Private Sub Setup_any() Implements Vixen.IEventDrivenOutputPlugIn.Setup
'		msgbox("TODO; just edit file for now by creating a copy of target plugins within VirtPlugins node")
		Try
'10:		Dim dialog As FxGenSetup = New VixFxGenSetup(Me.m_setupData, Me.m_setupNode, Me.m_numch, Me.m_rcvbuf)
'11:		Dim result As DialogResult = dialog.ShowDialog()
'12:		If result = DialogResult.OK Then
'#If False
'10:		Me.LoadConfig(False) 'get config info if not already loaded
'10:		Me.TimingTrack_seq() 'don't do this one until run-time (channel assignments might change)
11:		Me.LoadFrameTags_seq()
12:		If Logger.WantTrace Then Logger.LogMsg("Setup")
13:		Dim dlg As New VixFxGenSetup(Me) '= New VixFxGenSetup(Me.m_logger)
20:		dlg.LogFile = Me.VixenInfo.Abs2Rel(Logger.LogFile)
24:		dlg.DataLogging = Logger.WantTrace
22:		dlg.PauseOnError = Logger.ErrPause
23:		dlg.AudioControl = Me.m_audio
25:		dlg.ReadSeq = Me.m_rdseq
26:		dlg.WriteSeq = Me.m_wrseq
		dlg.Chofs = Me.m_startch
27:		If Me.m_virtseq IsNot Nothing Then
			dlg.VirtProfile = Me.VixenInfo.Abs2Rel(Me.m_virtseq.FileName) 'should be fully qualified path name
			dlg.MaxChofs = Me.VixenInfo.EndChannel - MaxChannel + 1 'max channel offset amount that will still fit within sequence
		Else
			dlg.MaxChofs = 1 'dummy value; no channels yet
		End If
#If False 'redundant
13:		For i As Integer = 1 To m_plugins.Count 'sorted by name
14:			If Logger.WantTrace Then Logger.LogMsg(String.Format("add plugin#{0}/{1} '{2}' to listbox", i, m_plugins.Count, m_plugins(i - 1).Desc)) '.Keys(i - 1)))
15:			dlg.PluginNames(i - 1) = m_plugins(i - 1).Desc '.Keys(i - 1)
16:			dlg.PluginSelected(i - 1) = m_plugins(i - 1).WantFx '.Enabled '.Values(i - 1).Selected '.VirtEnabled
'17:			dlg.PluginChannels(i - 1) = New Size(Me.m_plugins(i - 1).StartChannel, Me.m_plugins(i - 1).EndChannel)
			Try 'might have invalid residue from previous virt profile
				dlg.PixelType(i - 1) = m_plugins(i - 1).PixelType
				dlg.CanvasSize(i - 1) = Size2String(m_plugins(i - 1).Geometry) 'FxGenPluginBase.
				dlg.ChannelOrder(i - 1) = m_plugins(i - 1).ChannelOrder
				dlg.PluginTag(i - 1) = m_plugins(i - 1).PluginTag
			Catch exc As Exception
				Logger.ReportError("PluginSetup", exc)
			End try
		Next i
#End If
'21:		dlg.EffectsFile = Me.m_fxfile 'IIf((Me.m_fxmain IsNot Nothing) And (Me.m_chmapper IsNot Nothing), Me.m_fxfile, String.Empty)
'		Try
'30:			If Not String.IsNullOrEmpty(dlg.EffectsFile.Path) Then
'31:				If Not String.IsNullOrEmpty(Me.m_pixtype) Then dlg.PixelType = Me.m_pixtype
'32:				If Me.m_gridsize.Width * Me.m_gridsize.Height <> 0 Then dlg.CanvasSize = String.Format("{0} x {1}", Me.m_gridsize.Width, Me.m_gridsize.Height)
'33:				If Not String.IsNullOrEmpty(Me.m_chorder) Then dlg.ChannelOrder = Me.m_chorder
'			End If
'		Catch exc As Exception 'ignore errors
'34:			Me.m_logger.ReportError("Set channel mapper props", exc)
'		End Try
35:		If Logger.WantTrace Then Logger.LogMsg("setup dlg")', LogMode.Flush)
36:		Dim result As DialogResult = dlg.ShowDialog()
'		If result = DialogResult.Ignore Then result = DialogResult.None 'kludge: put it back to correct value
		If Logger.WantTrace Then Logger.LogMsg(String.Format("dlg result: {0}", result.ToString()))
37:		If result <> DialogResult.Yes Then 'Or Not dlg.IsDirty Then 'no changes; explicitly warn user that data was not saved
			If result <> DialogResult.None Or True Then MessageBox.Show(String.Format("NOT Saved: '{0}' was selected", result.ToString()), "NOT Saved", System.Windows.Forms.MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
'38:			If result <> DialogResult.None Then MessageBox.Show(String.Format("NOT saved: '{1}' was selected", TypeName(result), result.ToString(), TypeName(DialogResult.OK), DialogResult.OK), "NOT Saved", System.Windows.Forms.MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
39:			dlg.Dispose()
			Return
		End If
40:		Me.WriteConfig("LogFile", dlg.LogFile)
'41:		Me.WriteConfig("EffectsFile", dlg.EffectsFile.Path)
42:		Me.WriteConfig("PauseOnError", dlg.PauseOnError)
43:		Me.WriteConfig("AudioControl", dlg.AudioControl)
44:		Me.WriteConfig("DataLogging", dlg.DataLogging)
45:		Me.WriteConfig("ReadSequence", dlg.ReadSeq)
46:		Me.WriteConfig("WriteSequence", dlg.WriteSeq)
		Me.WriteConfig("StartChannel", dlg.Chofs)
		Me.WriteConfig("VirtSeq", dlg.VirtProfile)
		Me.WriteConfig("LastModified", Date.Now.ToString()) 'audit trail

		If Me.m_setupNode.SelectNodes("FrameTags/Tag").Count < 1 Then 'populate initial frame tags with func codes for sanity check
			If logger.WantTrace Then Logger.LogMsg("Write initial frame tags")
			Dim frametags As XmlNode = Vixen.Xml.SetValue(Me.m_setupNode, "FrameTags", String.Empty) 'NOTE: wipes out prev node
			For Each keyvalue As KeyValuePair(Of String, Integer) In Me.EnumFuncCodes()
				Dim frtag As XmlNode = Vixen.Xml.SetNewValue(frametags, "Tag", keyvalue.Key) 'function name
				Vixen.Xml.SetAttribute(frtag, "track", "fxfunc")
				Vixen.Xml.SetAttribute(frtag, "value", keyvalue.Value.ToString())
			Next keyvalue
		End If
'47:		Me.WriteConfig("PixelType", dlg.PixelType)
'48:		Me.WriteConfig("ChannelOrder", dlg.ChannelOrder)
'49:		Me.WriteConfig("CanvasSize", dlg.CanvasSize)

'30:		Me.m_logger.LogFile = dlg.LogFile
'31:		Me.m_logger.LogMsg("write cfg updates to seq/prof")
'32:		Me.m_fxfile = dlg.EffectsFile
'33:		Me.m_logger.ErrPause = dlg.PauseOnError
'34:		Me.m_audio = dlg.AudioControl
'35:		Me.m_logger.WantTrace = dlg.DataLogging
'36:		Me.m_wronly = dlg.[WriteOnly]
'37:		Me.m_pixtype = dlg.PixelType
'38:		Me.m_chorder = dlg.ChannelOrder
'39:		Me.m_gridsize = FxUtils.String2Size(dlg.CanvasSize)
50:		Dim virtPluginsNode As XmlNode = Vixen.Xml.SetValue(Me.m_setupNode, MyPluginsNode, String.Empty) 'NOTE: wipes out prev node
'51:		Dim sb As New StringBuilder()
52:		For i As Integer = 1 To m_plugins.Count
53:			If Not dlg.PluginSelected(i - 1) Then Continue For 'don't bother saving info unless fx enabled
'53:			If dlg.PluginSelected(i - 1) <> Me.m_plugins(i - 1).Enabled Then sb.AppendFormat("{0} has been {1}.{2}", Me.m_plugins(i - 1).Desc, IIf(dlg.PluginSelected(i - 1), "disabled for now", "re-enabled"), vbCRLF)
55:			m_plugins(i - 1).UpdateConfig(virtPluginsNode, dlg.PluginSelected(i - 1), dlg.PixelType(i - 1), String2Size(dlg.CanvasSize(i - 1)), dlg.ChannelOrder(i - 1), dlg.PluginTag(i - 1))
		Next i
'55:		Me.LoadConfig(True) 'force reload of config info
'		Me.m_plugins.UpdateCounts()
56:		dlg.Dispose()

#If False
'75:		If pluginsNode.ChildNodes.Count < 1 Then MessageBox.Show("No output plug-ins have been selected.", "No Plug-ins?", System.Windows.Forms.MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
60:		If sb.length > 0 Then
61:			If Form.ActiveForm IsNot Nothing Then
62:				Form.ActiveForm.Close()
			Else
63:				sb.Append("Please close and re-open the ""Sequence Plugins Mapping"" window to see updated status.")
			End If
64:			MessageBox.Show(sb.ToString(), "Plugin Status Updated", System.Windows.Forms.MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
		End If
#End If
65:		If Logger.WantTrace Then Logger.LogMsg("config save: success") ', Me.m_logger.WantTrace)
		MessageBox.Show("Changes saved", "Saved", System.Windows.Forms.MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
'#End If
'27:		Me.LoadChannels()
'26:		Me.LoadFrameTags()
'no66:		Me.TimingTrack()
		Me.LoadConfig(True)
		Catch exc As Exception
			Logger.ReportError("Setup", exc)
		End Try
	End Sub


'from Vixen tutoral:
'"Called when a sequence is executed. Returns a list of forms that the plug-in needs to have injected
'into Vixen's MDI interface. The previews are examples Of plug-ins that Return forms For this purpose."
'my additional comments:
'Vixen interface is broken (function prototype is different between Vixen 2.1 and 2.5), so a separate version-specific wrapper function will call this generic function
	Private Sub Startup_any() 'As List(Of Form) 'Implements Vixen2508.IHardwarePlugin.Startup, Vixen.IOutputPlugIn.Startup '2.1.1.0
		Try
'			Me.m_shuttingdown = False
10:		Me.TimingTrack_seq()
11:		Me.LoadFrameTags_seq()
'12:		If Not Me.m_plugins.LoadPlugins(Me.m_virtseq) Then Return 'nowhere to send output
'		If Me.m_framebuf.Length < Me.VixenInfo.EndChannel - Me.VixenInfo.StartChannel +1 Then Throw New Exception(String.Format("Frame output buffer (len {0}) is smaller than frame input buffer (len {1})", Me.m_framebuf.Length, Me.VixenInfo.EndChannel - Me.VixenInfo.StartChannel +1)) 'needed in order to use output as inbuf first time (lazy programmers)
'		Me.m_fx.FxGen(DirectCast(Me.VixenInfo.Seq, Vixen.EventSequence), 0, Me.m_framebuf) 'pre-render first frame so it's ready when needed
'		Me.VixenInfo.total_inlen += Me.VixenInfo.EndChannel - Me.VixenInfo.StartChannel +1 'channelValues.Length
'		Me.VixenInfo.total_evts += 1
13:		Me.FrameBuf = Me.AllocFramebuf() 'New Byte(PluginWrapper.MaxChannel - PluginWrapper.MinChannel +2) {} 'allow trailing B + G on RGB values
14:		Me.StartAllPlugins() 'Me.m_virtseq)
15:		Me.m_rendered = -1 'no prev frame rendered
		Me.FxRender(Integer.MinValue, Integer.MinValue, Nothing) 'initialize fx
		Me.m_fxDataChanged = False 'no data yet
16:		Dim initframe As Byte() = New Byte(Me.VixenInfo.EndChannel - Me.VixenInfo.StartChannel +0) {}
17:		Me.Event_any(initframe) 'pre-render first frame so it's ready when needed
18:		Me.m_fxDataChanged = True 'force refresh first time
'		dim gdoor_mapper as new
		Catch exc As Exception
			Logger.ReportError("Startup", exc)
			Me.m_wkerstate = WkerState.ErrorOccurred
'?			Me.Shutdown_any()
		End Try
	End Sub

'from Vixen tutoral:
'"Called when the plug-in needs to update the hardware state during execution.
'	channelValues: Event values in channel order, 1 byte per channel."
'http://www.doityourselfchristmas.com/forums/showthread.php?p=37947
'"The event method is called whenever there is a change in the state of the channel values. With each call, the method receives
'the data For every channel that's being sent so that plugin."
'my additional comments:
'Most of the processing during the sequence occurs within this event handler.
'A background thread is used to avoid missing events on Vixen's main thread.
'	Private m_channelVals As Byte() = Nothing 'pass channel vals from main Vixen thread to child thread
'	Private m_svchannelVals As Byte() = Nothing 'channel vals from prev frame
'	Private m_quelen As New Dictionary(Of Integer, Integer) 'queue stats
'	Private m_frameque As New List(Of Integer) 'list of frame#s waiting for bkg processing
'	Private m_prevframe As Integer 'last-processed frame#
	Private Sub Event_any(channelValues As Byte()) Implements Vixen.IEventDrivenOutputPlugIn.[Event]
		Try
'			If Me.m_shuttingdown Then msgbox("ignore trailing event"): Return
40:			Dim timestamp As Integer = &H10000 * channelValues(Me.VixenInfo.TimingTrack - Me.VixenInfo.StartChannel +1 + 0) + &H100 * channelValues(Me.VixenInfo.TimingTrack - Me.VixenInfo.StartChannel +1 + 1) + channelValues(Me.VixenInfo.TimingTrack - Me.VixenInfo.StartChannel +1 + 2)
41:			Me.VixenInfo.CurrentFrame = CInt(timestamp/DirectCast(Me.VixenInfo.Seq, Vixen.EventSequence).EventPeriod) '0-based (before frame), 1-based (after frame)
42:			Me.m_wkerstate = WkerState.NextFrame
'speculative pre-render will only be incorrect for loops or subrange playback; this is typically testing, so timing is not as critical
'	Logger.LogMsg(String.Format("evt @{0} msec: pre-rendered {1}, new frame {2}, changed? {3}", timestamp - DirectCast(Me.VixenInfo.Seq, Vixen.EventSequence).EventPeriod, Me.m_rendered, Me.VixenInfo.CurrentFrame - 1, Me.m_fxDataChanged))
43:			Me.FxRender(Me.VixenInfo.CurrentFrame - 1, timestamp - DirectCast(Me.VixenInfo.Seq, Vixen.EventSequence).EventPeriod, channelValues) 'render frame now if it was not pre-rendered
44:			If Me.m_fxDataChanged Then 'pre-rendered, channels need to be updated
45:				For i As Integer = 1 To Me.FrameBuf.Length - 2 'CAUTION: expensive loop just for stats
46:					If Me.FrameBuf(i -1) = 0 Then Me.VixenInfo.total_nullout += 1
				Next i
47:				SendAllPlugins(Me.FrameBuf) 'send pre-generated channel data for this frame
48:				Me.VixenInfo.total_outlen += Me.FrameBuf.Length - 2 '2 were extra for RGB overrun sloppyness
49:				Me.VixenInfo.total_macros += 1
			End If
'			If channelValues Is Nothing Then Return 'eof
50:			Me.m_fxDataChanged = False 'reset for next time
'			If Me.VixenInfo.CurrentFrame < 1 Then Return 'eof; no data to pre-render for next time
51:			If (Me.VixenInfo.CurrentFrame < 0) Or (Me.VixenInfo.CurrentFrame >= DirectCast(Me.VixenInfo.Seq, Vixen.EventSequence).TotalEventPeriods) Then Return 'eof; no data to pre-render for next time
'now generate data for *next* frame:
'this generates one frame ahead in order to avoid latency (heavy processing is done after each event, not before)
'44:		If newframe >= Me.VixenInfo.Seq.TotalEventPeriods Then Me.m_wkerstate = WkerState.Shutdown: Return 'past eof; no further frame changes needed (0-based)
52:			Me.FxRender(Me.VixenInfo.CurrentFrame, timestamp, channelValues) 'pre-render for next time
53:			Me.VixenInfo.total_inlen += channelValues.Length
54:			Me.VixenInfo.total_evts += 1
55:			Windows.Forms.Application.DoEvents()
		Catch exc As Exception
			Logger.ReportError("Event", exc)
			Me.m_wkerstate = WkerState.ErrorOccurred
'?			Me.Shutdown_any()
		End Try
	End Sub

'from Vixen tutoral:
'"Called when execution is stopped or the plug-in instance is no longer going To be referenced."
	Private Sub Shutdown_any() Implements Vixen.IEventDrivenOutputPlugIn.Shutdown
		Try
'			Me.m_shuttingdown = True
'generate dummy framebuf for end of sequence:
'do we really need to do this?
'		If Me.m_fxDataChanged Then 'flush final frame
'			Dim finalframe As Byte() = New Byte(Me.VixenInfo.EndChannel - Me.VixenInfo.StartChannel +0) {}
''			Dim finalstamp As Integer = DirectCast(Me.VixenInfo.Seq, Vixen.EventSequence).TotalEventPeriods * DirectCast(Me.VixenInfo.Seq, Vixen.EventSequence).EventPeriod
''			finalframe(Me.VixenInfo.TimingTrack + 0) = CByte(finalstamp \ &H10000)
''			finalframe(Me.VixenInfo.TimingTrack + 1) = CByte((finalstamp \ &H100) Mod &H100)
''			finalframe(Me.VixenInfo.TimingTrack + 2) = CByte(finalstamp Mod &H100)
''	Logger.LogMsg(String.Format("final timestamp = {0} vs. {1}", &H10000 * finalframe(Me.VixenInfo.TimingTrack - Me.VixenInfo.StartChannel +1 + 0) + &H100 * finalframe(Me.VixenInfo.TimingTrack - Me.VixenInfo.StartChannel +1 + 1) + finalframe(Me.VixenInfo.TimingTrack - Me.VixenInfo.StartChannel +1 + 2), finalstamp))
''			Me.VixenInfo.CurrentFrame = Me.m_rendered + 1 'avoid re-rendering last frame, regardless of which one it was
'			Me.m_rendered = -1 'DirectCast(Me.VixenInfo.Seq, Vixen.EventSequence).TotalEventPeriods - 1 'make it look like final frame even if it isn't
'			Me.Event_any(finalframe)
'		End If
'20:		Me.m_wkerstate = WkerState.Shutdown 'TODO: move this?
		try
			Me.FxRender(Integer.MaxValue, Integer.MaxValue, Nothing) 'finalize fx
			StopAllPlugins() 'DirectCast(Me.VixenInfo.Seq, Vixen.EventSequence))
		Catch exc As Exception 'ignore extraneous collection problem
			Logger.ReportError("plugin stop", exc)
			If Logger.WantTrace Then Logger.LogMsg(String.Format("whoops: {0}", exc.ToString))
		end try
20:		Me.m_wkerstate = WkerState.Shutdown 'TODO: move this?
23:		If Logger.WantTrace Then Logger.LogMsg(String.Format("========== Shutdown at {0:MM-dd-yyyy HH:mm:ss}.{1:D3} ==========", New Object() { DateTime.Now, DateTime.Now.TimeOfDay.Milliseconds}))
24:		If Logger.WantTrace Then Logger.LogMsg("FxGen stats:")
25:		If Logger.WantTrace Then Logger.LogMsg(String.Format("Total events {0}, effects applied {1}, real bytes in {2} ({3}/frame), virtual bytes out {4} ({5}/frame), de-duped bytes out {6} ({7}/uniq frame), virtual compression (virt out/real in) {8}%, null virtual out bytes {9} ({10}% of virt out)", Me.VixenInfo.total_evts, Me.VixenInfo.total_macros, Me.VixenInfo.total_inlen, SafeDiv(Me.VixenInfo.total_inlen, Me.VixenInfo.total_evts), Me.VixenInfo.total_evts * (Me.FrameBuf.Length - 2), (Me.FrameBuf.Length - 2), Me.VixenInfo.total_outlen, SafeDiv(Me.VixenInfo.total_outlen, Me.VixenInfo.total_macros), 100 - SafeDiv(100 * Me.VixenInfo.total_inlen, Me.VixenInfo.total_evts * (Me.FrameBuf.Length - 2)), Me.VixenInfo.total_nullout, SafeDiv(100 * Me.VixenInfo.total_nullout, Me.VixenInfo.total_outlen)))', LogMode.Flush)
31:		Me.m_wkerstate = WkerState.AllDone
		Catch exc As Exception
			Logger.ReportError("Shutdown", exc)
			Me.m_wkerstate = WkerState.ErrorOccurred
		End Try
	End Sub
#End Region

#Region "Child helpers"
	'NOTE: .NET isn't allowing override of static, so this must go directly in CustomFx
'	Public MustOverride Function FxFuncChannel(plugtag As String) As Integer
	'return list of plugin tags indexed by fx func channel#:
'	Public Shared Function EnumFxProps(opnode As XmlNode) As Dictionary(Of Integer, String)
'		Dim list As New Dictionary(Of Integer, String)
'		For Each tagnode As XmlNode In opnode.SelectNodes("VirtPlugins/MappedPlugin/Tag")
'			Dim fxchfunc as Integer = FxFuncChannel(tagnode.InnerText)
'			If fxchfunc <> 0 Then list.Add(fxchfunc, tagnode.InnerText)
'		Next tagnode
'		Return list
'	End Function

	'return list of func or text frame tags:
	Public Shared Function EnumFrameTags(opnode As XmlNode, track As String) As Dictionary(Of String, Byte)
		'If logger.WantTrace Then Logger.LogMsg(String.Format("enum frame tags: track {0}, node {1}", track, opnode.OuterXml))
		Dim frtaglist As New Dictionary(Of String, Byte)
		For Each tagnode As XmlNode In opnode.SelectNodes(String.Format("FrameTags/Tag[@track='{0}']", track))
'TODO: handle timed tags
			If logger.WantTrace Then Logger.LogMsg(String.Format("got tag {0}", tagnode.OuterXml))
			If tagnode.Attributes("value") Is Nothing Then Continue For
			Dim val As Integer = Integer.Parse("0" & tagnode.Attributes("value").Value) 'default value to 0 if missing
			If (val > 0) And (val < 256) Then frtaglist.Add(tagnode.InnerText, CByte(val))
		Next tagnode
		If logger.WantTrace Then Logger.LogMsg(String.Format("returning {0} entrs", frtaglist.Count))
		Return frtaglist
	End Function
#End Region

#Region "Implementation/helpers"
'allow playback processing to run in a separate thread in case heavy fx processing bogs down main Vixen thread:
'TODO: spread this across multiple threads/cores?
'For now, just use multiple FxGen instances, so there's only one worker thread for each instance.
'That will avoid the need to control the order of updates to channels if mult fx funcs update same channels.
#If False
	Private Sub PlaybackThread()
'		Try
		Me.Startup_any()
'		Dim nxtframe As Byte() = Me.VixenFrame(1) 'any non-0 frame# tells bkg thread to process next frame
		While Me.m_wkerstate = WkerState.NextFrame
			Me.Event_any(nxtframe)
		End While
		Me.Shutdown_any()
'		Catch exc As Exception
'			Me.childstate = Error
'		End Try
	End Sub
#End If

'generate an array of channel bytes that looks like it came from Vixen:
'	Private Sub VixenFrame(wantframe As Integer) 'As Byte()
'		Dim framebuf As Byte() = New Byte(Me.VixenInfo.EndChannel - Me.VixenInfo.StartChannel + 1 -1) {}
'	me.m_logger.LogMsg(String.Format("alloc chbuf len {0}, ttrk ofs {1}-{2}+1={3}", framebuf.Length, Me.VixenInfo.TimingTrack, Me.VixenInfo.StartChannel, Me.VixenInfo.TimingTrack - Me.VixenInfo.StartChannel + 1), true)
'		Me.m_framebuf(Me.VixenInfo.TimingTrack - Me.VixenInfo.StartChannel +1 + 0) = CByte(wantframe \ &H10000)
'		Me.m_framebuf(Me.VixenInfo.TimingTrack - Me.VixenInfo.StartChannel +1 + 1) = CByte((wantframe \ &H100) Mod &H100)
'		Me.m_framebuf(Me.VixenInfo.TimingTrack - Me.VixenInfo.StartChannel +1 + 2) = CByte(wantframe Mod &H100)
'		Return framebuf
'	End Sub 'Function

	Private Sub FxRender(frame As Integer, timestamp As Integer, channelValues As Byte())
		If Me.m_rendered = frame Then Return 'already rendered this frame
		Dim regen_start As Long = Logger.Elapsed()
'46:		Me.GetFrameTags(newframe) 'get tags for next frame so fx funcs can act on them
'47:		If Me.m_rdseq Then Me.GetFrameChannels(newframe) 'start with sequence values before applying fx
'#If False 'need to comment this out for initial compile
#If False
		Me.m_fxDataChanged = False 'Me.FxGen(DirectCast(Me.VixenInfo.Seq, Vixen.EventSequence), Me.VixenInfo.CurrentFrame, timestamp, Me.VixenInfo.StartChannel -1, Me.m_framebuf) '(frame And 1)) 'render current frame
		Dim outbuf As New ByteRange(Me.FrameBuf, 0, 0) 'NOTE: this allows fx funcs to see entire frame buf if they want to look outside their own piece
		For Each vopiw As PluginWrapper In Me.m_plugins
			outbuf.StartOfs = vopiw.StartChannel -1
			outbuf.AvailLen = vopiw.EndChannel - vopiw.StartChannel +1
'//			this.m_canvas = this.m_chmap.Map(frbuf); //GDI bitmap to draw on
			Me.m_fxDataChanged = Me.m_fxDataChanged Or Me.FxGen(DirectCast(Me.VixenInfo.Seq, Vixen.EventSequence), frame, timestamp, vopiw.StartChannel - 1, outbuf) 'render current frame
			this.m_chmap.Unmap(this.m_canvas, frbuf); //refresh prop channel values from GDI canvas
		Next vopiw
#End If
		Me.m_fxDataChanged = RenderAllEffects(frame, timestamp) ', DirectCast(Me.VixenInfo.Seq, Vixen.EventSequence), frame, timestamp, Me.m_framebuf)
		Me.m_rendered = frame 'remember which frame is rendered so we don't do it again
		If Logger.WantTrace And (channelValues IsNot Nothing) Then Logger.LogMsg(String.Format("evt @{0} msec: new data? {1}, fxgen took {2} msec, cur frame {3}, {4} channel bytes in: {5}", timestamp, Me.m_fxDataChanged, Logger.Elapsed() - regen_start, frame, channelValues.Length, Logger.DumpHex(channelValues, 0, channelValues.Length, 200)))
		Me.SaveFrameChannels(frame) 'save data if user wants it
'#End If
	End Sub

'load channel-related info:
'assume this info won't change during the plug-in lifespan, so it can be loaded once during the Init event
	Private Sub LoadChannels()
'		Me.m_stchannel = Convert.ToInt32(Me.m_setupNode.Attributes.ItemOf("from").Value) - 1 '0-based
		Me.VixenInfo.Channels = Me.VixenInfo.Seq.Channels 'used to generate timing track + macro functions
		Me.VixenInfo.StartChannel = Integer.Parse("0" & Me.m_setupNode.Attributes("from").InnerText) '1-based
		Me.VixenInfo.EndChannel = Integer.Parse("0" & Me.m_setupNode.Attributes("to").InnerText) '1-based
		Me.VixenInfo.TimingTrack = Me.VixenInfo.EndChannel - 3 'use upper 3 channels for timing track; 0-based on full channel array from Vixen
		Dim numch As Integer = Me.VixenInfo.EndChannel - Me.VixenInfo.StartChannel +1
		If Logger.WantTrace Then Logger.LogMsg(String.Format("LoadChannels {0}..{1} ({2} channels), timing at {3}..{4}", Me.VixenInfo.StartChannel, Me.VixenInfo.EndChannel, numch, Me.VixenInfo.TimingTrack + 1, Me.VixenInfo.TimingTrack + 3))
'		Me.TimingTrack()
	End Sub

'generate timing track (3 channels):
'At 1 msec interval, 24 bits allows 16M msec ~= 4.7 hrs of run time (25 msec would be 116 hr).
'The timing track will force Vixen to call the plug-in for every frame, and will also tell the plug-in which frame (in case Vixen is skipping or looping frames).
'However, Vixen seems to filter out small channel changes, so the top bit is toggled to force a larger change.
	Private Sub TimingTrack_seq() 'can only be called on a sequence, not a profile
		Dim numch As Integer = Me.VixenInfo.EndChannel - Me.VixenInfo.StartChannel +1
		if numch < 3 Then Throw New Exception(String.Format("FxGen needs at least 3 channels; currently assigned: {0}..{1} ({2})", Me.VixenInfo.StartChannel, Me.VixenInfo.EndChannel, numch))
		Dim wrtiming As Boolean = False
		Dim evtvals As Byte(,) = DirectCast(Me.VixenInfo.Seq, Vixen.EventSequence).EventValues
		If Logger.WantTrace Then Logger.LogMsg(String.Format("channels in TimingTrack: {0}x{1} vs. {2}x{3}, timing ch {4}..{5}", evtvals.GetUpperBound(0), evtvals.GetUpperBound(1), DirectCast(Me.VixenInfo.Seq, Vixen.EventSequence).EventValues.GetUpperBound(0), DirectCast(Me.VixenInfo.Seq, Vixen.EventSequence).EventValues.GetUpperBound(1), Me.VixenInfo.TimingTrack, Me.VixenInfo.TimingTrack + 2))
'		Dim ttkofs As Integer = Me.VixenInfo.EndChannel - 3 'timing track channel offset
'NOTE: can only be called on a sequence, not a profile
		For frame As Integer = 1 To DirectCast(Me.VixenInfo.Seq, Vixen.EventSequence).TotalEventPeriods
			Dim want_timestamp As Integer = frame * DirectCast(Me.VixenInfo.Seq, Vixen.EventSequence).EventPeriod
			If Not wrtiming Then
				Dim got_timestamp As Integer = &H10000 * evtvals(Me.VixenInfo.TimingTrack + 0, frame - 1 + 0) + &H100 * evtvals(Me.VixenInfo.TimingTrack + 1, frame - 1) + evtvals(Me.VixenInfo.TimingTrack + 2, frame - 1)
				If got_timestamp = want_timestamp Then Continue For
				Dim mchannels As String = String.Format(IIf(Me.VixenInfo.StartChannel < Me.VixenInfo.TimingTrack +1, "{0} - {1}", "{0}").ToString(), Me.VixenInfo.StartChannel, Me.VixenInfo.TimingTrack +1)
				Dim response As DialogResult = MessageBox.Show(String.Format("Timing track is missing or damaged, and will be written to channels {0} - {1} (found {2} instead of {3} at frame# {4}).  Do you also want to set colors/labels for Macro channel {5}?", Me.VixenInfo.TimingTrack + 1, Me.VixenInfo.TimingTrack + 3, got_timestamp, want_timestamp, frame, mchannels), Me.Name_any, System.Windows.Forms.MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question)
				If Logger.WantTrace Then Logger.LogMsg(String.Format("Timing track not found in channels {0} - {1} (found {2} instead of {3} at frame# {4}).  User response: {5} = write timing? {6}, write labels? {7}", Me.VixenInfo.TimingTrack + 1, Me.VixenInfo.TimingTrack + 3, got_timestamp, want_timestamp, frame, response, response <> DialogResult.Cancel, response = DialogResult.Yes))
				If response = DialogResult.Cancel Then Throw New Exception("FxGen timing track setup cancelled.") 'exc is the only way to tell Vixen to abort the Setup
				wrtiming = True
				If response = DialogResult.Yes Then
					Dim chnames As Dictionary(Of Integer, ColoredString) = GetPluginFxChannels(Vixen.Xml.GetNodeAlways(Me.m_setupNode, MyPluginsNode, String.Empty))
'set colors and labels for non-timing tracks:
					For ch As Integer = Me.VixenInfo.StartChannel To Me.VixenInfo.EndChannel
						If chnames.ContainsKey(ch) Then
							Me.VixenInfo.Channels(ch - 1).Name = chnames(ch).text
							Me.VixenInfo.Channels(ch - 1).Color = chnames(ch).color
						Else
							Me.VixenInfo.Channels(ch - 1).Name = String.Format("FxGen Macro[{0}] #{1}", ch - Me.VixenInfo.StartChannel, ch)
							Me.VixenInfo.Channels(ch - 1).Color = System.Drawing.Color.FromArgb(255, &HC0, &HF0, &HFF) 'light blue
						End If
						If Logger.WantTrace Then Logger.LogMsg(String.Format("set ch {0} name to {1}", ch - 1, Me.VixenInfo.Channels(ch - 1).Name))
					Next ch
'set timing track channel colors and labels:
					For ch As Integer = Me.VixenInfo.TimingTrack + 0 To Me.VixenInfo.TimingTrack + 2
						Me.VixenInfo.Channels(ch).Color = System.Drawing.Color.FromArgb(255, &HA0, &HA0, &HA0) 'light gray
						Me.VixenInfo.Channels(ch).Name = String.Format("FxGen timing[{0}] #{1}", ch - Me.VixenInfo.TimingTrack, ch + 1)
					Next ch
				End If
			End If
			'timing track tells plug-in which frame is playing, and forces Vixen to call FxGen for each frame
			evtvals(Me.VixenInfo.TimingTrack + 0, frame - 1) = CByte(want_timestamp \ &H10000)
			evtvals(Me.VixenInfo.TimingTrack + 1, frame - 1) = CByte((want_timestamp \ &H100) Mod &H100)
			evtvals(Me.VixenInfo.TimingTrack + 2, frame - 1) = CByte(want_timestamp Mod &H100)
		Next frame
	End Sub

'load frame tags:
'assume this info won't change during the plug-in lifespan, so it can be loaded once during the Init event
'TODO: this will be for cue-based effect triggering eventually; currently does nothing
	Private Sub LoadFrameTags_seq() 'can only be called on a sequence, not a profile
		Try
		If Logger.WantTrace Then Logger.LogMsg(String.Format("LoadFrameTags: seq len {0}", DirectCast(Me.VixenInfo.Seq, Vixen.EventSequence).TotalEventPeriods))
'		msgbox("seq pid? " & (Me.VixenInfo.Seq.PlugInData isnot nothing) & ", exec pid? " & (me.VixenInfo.Exec.PlugInData isnot nothing))
'		Dim numtags As Integer = 0
'		Dim trackused As New Dictionary(Of String, Integer) 'remember last used frame# per track
'look for frame tags in any plug-in; TODO: decide which plug-in should be the owner?
		FrameTag.Clear()
10:		For Each vop As Vixen.OutputPlugin In Me.VixenInfo.Seq.PlugInData.GetOutputPlugins()
'			If vop.Name <> plugin.Name Then Continue For
11:			Dim opnode As XmlNode = Me.VixenInfo.Seq.PlugInData.GetPlugInData(vop.Id.ToString)
12:			For Each tagnode As XmlNode In opnode.SelectNodes("FrameTags/Tag")
'				Dim trk As String = SafeValue(tagnode.Attributes("track"), "none") 'default track name is missing
'				If Not trackused.ContainsKey(trk) Then trackused.Add(trk, 0) 'default track start is beginning of seq
'13:				Dim stframe As Integer = Me.Time2Frame_seq(safevalue(tagnode.Attributes("start"), trackfrom(trk))) 'default tag start
'14:				Dim enframe As Integer = Me.Time2Frame_seq(tagnode.Attributes("end").Value)
'15:				Dim track As String = tagnode.Attributes("track").Value 'tracks can be Papagayo "voices"
'16:				Dim frtag As FxGen.FrameTag = New FxGen.FrameTag(track, tagnode.InnerText, stframe, enframe)
16:				Dim frtag As FxGen.FrameTag = New FxGen.FrameTag(tagnode, Me.VixenInfo)
'				Logger.LogMsg(String.Format("load frametag, start {0}, end {1}, track {2}, tag {3}", stframe, enframe, track, tagnode.InnerText))
'17:				If Not Me.m_frtags.ContainsKey(stframe) Then Me.m_frtags.Add(stframe, New List (Of FxGen.FrameTag))
'18:				If Not Me.m_frtags.ContainsKey(enframe) Then Me.m_frtags.Add(enframe, New List (Of FxGen.FrameTag))
'19:				Me.m_frtags(stframe).Add(frtag)
'20:				Me.m_frtags(enframe).Add(frtag) 'indexed by start and end frame
'21:				numtags += 1
			Next tagnode
		Next vop
'test
'		If Not Me.m_notes.ContainsKey(0) Then Me.m_notes(0) = "START"
'		If Not Me.m_notes.ContainsKey(20) Then Me.m_notes(20) = "     ONE SEC"
'		If Not Me.m_notes.ContainsKey(200) Then Me.m_notes(200) = "TEN SEC"
22:		If Logger.WantTrace Then Logger.LogMsg(String.Format("#tracks: {0}, #frame tags loaded: {1}", FrameTag.TrackCount, FrameTag.TagCount))
		Catch exc As Exception
			Logger.ReportError("LoadFrameTags", exc)
		End Try
	End Sub


'load config data:
'this info can change after Setup, so it may be reloaded
	Private Sub LoadConfig(force As Boolean)
		If Logger.WantTrace Then Logger.LogMsg(String.Format("LoadConfig(force {0})", force))
		If Me.m_cfgok And Not force Then Return 'already loaded
		Me.m_cfgok = False
		Try
'load general config:
10:		Logger.LogFile = Me.VixenInfo.Rel2Abs(Me.VixenInfo.SetupData.GetString(Me.m_setupNode, "LogFile", String.Empty))
12:		Logger.ErrPause = Me.VixenInfo.SetupData.GetBoolean(Me.m_setupNode, "PauseOnError", False)
14:		Logger.WantTrace = Me.VixenInfo.SetupData.GetBoolean(Me.m_setupNode, "DataLogging", False)
'11:		Me.m_fxfile.path = Me.VixenInfo.SetupData.GetString(Me.m_setupNode, "EffectsFile", String.Empty)
13:		Me.m_audio = Me.VixenInfo.SetupData.GetBoolean(Me.m_setupNode, "AudioControl", False)
15:		Me.m_rdseq = Me.VixenInfo.SetupData.GetBoolean(Me.m_setupNode, "ReadSequence", False)
		Me.m_startch = Me.VixenInfo.SetupData.GetInteger(Me.m_setupNode, "StartChannel", 0)
16:		Me.m_wrseq = Me.VixenInfo.SetupData.GetBoolean(Me.m_setupNode, "WriteSequence", False)
21:		Dim virtfile As String = Me.VixenInfo.Rel2Abs(Me.VixenInfo.SetupData.GetString(Me.m_setupNode, "VirtSeq", String.Empty))
'17:		Me.m_pixtype = Me.VixenInfo.SetupData.GetString(Me.m_setupNode, "PixelType", String.Empty)
'18:		Me.m_gridsize = FxUtils.String2Size(Me.VixenInfo.SetupData.GetString(Me.m_setupNode, "CanvasSize", "0 x 0"))
'19:		Me.m_chorder = Me.VixenInfo.SetupData.GetString(Me.m_setupNode, "ChannelOrder", String.Empty)
20:		If Logger.WantTrace Then Logger.LogMsg(String.Format("config: PauseOnError {0}, AudioControl {1}, DataLogging {2}, ReadSequence {3}, WriteSequence {4}, ChannelOffset {5}, VirtProfile {6}", Logger.ErrPause, Me.m_audio, Logger.WantTrace, Me.m_rdseq, Me.m_wrseq, Me.m_startch, virtfile)) 'Me.m_pixtype, Me.m_gridsize.Width, Me.m_gridsize.Height, Me.m_chorder))
'Me.m_logger.WantTrace = True

#If False
'load plug-in selection:
30:		For Each plug As PlugInfo In Me.m_plugins.Values
			plug.VirtEnabled = False
31:			Dim fxnode As XmlNode = Me.m_setupNode.SelectSingleNode(String.Format("VirtPlugins/Plugin[@id='{0}']", plug.ID)) '(0)
			If fxnode Is Nothing Then Continue For
'32:			plug.Selected = (fxnode IsNot Nothing) AndAlso CBool(fxnode.Attributes("enabled").Value)
'			If plug.Selected Then
32:			plug.VirtEnabled = CBool(fxnode.Attributes("enabled").Value)
			plug.VirtStartChannel = CInt(fxnode.Attributes("from").Value) 'Me.VixenInfo.SetupData.GetInteger(fxnode, "VirtualStartChannel", plug.RealStartChannel)
			plug.VirtEndChannel = CInt(fxnode.Attributes("to").Value) 'Me.VixenInfo.SetupData.GetInteger(fxnode, "VirtualEndChannel", plug.RealEndChannel)
'			End If
33:			If Me.m_logger.WantTrace Then Me.m_logger.LogMsg(string.Format("plug {0} found in setup node {1}? {2}", plug.Desc, String.Format("VirtPlugins/Plugin[id='{0}']", plug.ID), plug.Selected))
		Next plug
34:		Me.m_plugins.UpdateCounts()
35:		Me.FrameBuf = Nothing 'force realloc in case #channels changed
'36:		If Me.m_fxfile.chmapper IsNot Nothing Then
'			Try
''37:				Me.m_fxfile.chmapper.Logger = Me.m_logger
'38:				Me.m_fxfile.chmapper.FxEvt.NumFrames = Me.VixenInfo.Seq.TotalEventPeriods
'39:				Me.m_fxfile.chmapper.FxEvt.FrameTags = String.Empty
'40:				Me.m_fxfile.chmapper.PixelType = Me.m_pixtype
'41:				Me.m_fxfile.chmapper.Geometry = Me.m_gridsize
'42:				Me.m_fxfile.chmapper.ChannelOrder = Me.m_chorder
'			Catch exc As Exception
'				Me.m_logger.ReportError("Set up channel mapping", exc)
'			End Try
'		End If
#End If
22:		If Not String.IsNullOrEmpty(virtfile) Then
			LoadVirtProfile(virtfile)
		ElseIf force Then
			Throw New Exception("No virtual sequence defined for FxGen")
		End If
		Catch exc As Exception
			Logger.ReportError("LoadConfig", exc)
		End Try
	End Sub

	Private Const MyPluginsNode As String = "VirtPlugins"
	Public Sub LoadVirtProfile(virtfile As String)
		Dim tried As New StringBuilder()
'		Dim virtnode As XmlNode = Me.m_setupNode.SelectSingleNode(MyPluginsNode)
'		If virtnode Is Nothing Then virtnode = Vixen.Xml.SetValue(Me.m_setupNode, MyPluginsNode, String.Empty)
		Dim virtnode As XmlNode = Vixen.Xml.GetNodeAlways(Me.m_setupNode, MyPluginsNode, String.Empty)
23:		Dim folders() As String = {String.Empty, System.IO.Path.GetDirectoryName(Me.VixenInfo.Seq.FileName), System.Environment.CurrentDirectory, System.IO.Path.GetDirectoryName(Process.GetCurrentProcess().MainModule.FileName)} 'allow relative path
		If Logger.WantTrace Then Logger.LogMsg(String.Format("load virt profile {0}", virtfile))
		For where As Integer = 1 To folders.Length
			If InStr(tried.ToString() & ",", String.Format(", {0},", folders(where -1))) <> 0 Then Continue For 'already tried this one
24:			tried.AppendFormat(", {0}", folders(where -1))
25:			Dim path As String = System.IO.Path.Combine(folders(where -1), virtfile)
'			Logger.LogMsg(String.Format("'{0}' found? {1}", path, System.IO.File.Exists(path)))
26:			If Not System.IO.File.Exists(path) Then Continue For
			If System.IO.Path.GetFileName(path) = Me.VixenInfo.Seq.FileName Then Throw New Exception("Recursive virtual sequences are not supported!")
			Try
27:				Me.m_virtseq = New Vixen.Profile(path)
'				Logger.LogMsg(String.Format("loaded virtual (nested) profile '{0}'", path))
				Me.m_cfgok = True
				If Me.m_virtseq IsNot Nothing Then LoadAllPlugins(Me.m_virtseq, virtnode)
				Return
			Catch exc As Exception
				If Logger.WantTrace Then Logger.LogMsg(String.Format("{0} is not a profile: {1}", path, exc.Message)) '.ToString()))
			End Try
			Try
28:				Me.m_virtseq = New Vixen.EventSequence(path)
'				Logger.LogMsg(String.Format("loaded virtual (nested) sequence '{0}'", path))
				Me.m_cfgok = True
				If Me.m_virtseq IsNot Nothing Then LoadAllPlugins(Me.m_virtseq, virtnode)
				Return
			Catch exc As Exception
				If Logger.WantTrace Then Logger.LogMsg(String.Format("{0} is not a sequence: {1}", path, exc.Message)) '.ToString()))
			End Try
		Next where
		Throw New Exception(String.Format("virt seq {0} not found, tried {1}", virtfile, tried.ToString().Substring(2)))
	End Sub

'get tags for this frame
'	Private Sub GetFrameTags(framenum As Integer)
'		If Not Me.m_frtags.ContainsKey(framenum) Then Return 'no tag state changes for this frame
'		For Each frtag As FxGen.FrameTag In Me.m_frtags(framenum)
'			If frtag.StartFrame = framenum Then Me.m_acttags.Add(frtag.Text) 'show tag
'			If frtag.EndFrame = framenum Then Me.m_acttags.Remove(frtag.Text) 'hide tag
'		Next frtag
'		Me.VixenInfo.FrameTags = FxUtils.JoinString(", ", Me.m_acttags.ToArray())
'		If Logger.WantTrace Then Logger.LogMsg(String.Format("frame tags: {0}", Me.VixenInfo.FrameTags))
'	End Sub


'read channel values for a frame from sequence:
'also allocates channel buffer to pass channel values to output plug-ins
'Since Adjustable Preview and Grid Editor typically overlap channels with other plug-ins,
'one large frame buffer is allocated rather than a smaller buffer for each plug-in.
'This allows only one read or write cycle to be used (for hopefully better performance).
#If False 'fx child can't see this, so never mind
	Public Function GetFrameChannels(framenum As Integer, startch as Integer, numch as Integer) As Byte()
'		Dim numch As Integer = Me.m_plugins.MaxChannel - Me.m_plugins.MinChannel + 1
'		If (Me.m_framebuf Is Nothing) OrElse (Me.m_framebuf.Length < numch) Then
'			Me.m_framebuf = New Byte(numch -1) {}
'			If Logger.WantTrace Then Logger.LogMsg(String.Format("allocated frame buffer for {0} channels {1}..{2}", Me.m_framebuf.Length, Me.m_plugins.MinChannel, Me.m_plugins.MaxChannel))
'		End If
'		If Not Me.m_rdseq Then Return 'don't need to read values from sequence
		Dim endch As Integer = Math.Min(DirectCast(Me.VixenInfo.Seq, Vixen.EventSequence).ChannelCount, startch + numch -1)
		If Logger.WantTrace Then Logger.LogMsg(String.Format("read {0} channels {1}..{2} from seq frame {3}", numch, startch, endch, framenum))
		If endch < startch Then Return New Byte(-1) {}
		Dim retval As Byte() = New Byte(endch - startch) {}
		For ch As Integer = startch To endch
			retval(ch - startch) = DirectCast(Me.VixenInfo.Seq, Vixen.EventSequence).EventValues(ch - startch, framenum) 'frame# is 0-based, channel#s are 1-based
		Next ch
'		Windows.Forms.Application.DoEvents()
		Return retval
	End Function

	Public Function GetFrameChannel(ch As Integer, abs As Boolean) As Byte
		If Not abs Then ch += Me.VixenInfo.StartChannel -1 + Me.m_startch - 1 'relative to my channel range
		Return DirectCast(Me.VixenInfo.Seq, Vixen.EventSequence).EventValues(ch -1, Me.VixenInfo.CurrentFrame)
	End Function
#End If

'write channels values back to sequence:
'This allows the effects to be played back when FxGen is disabled, to eliminate CPU bottleneck on slower machines.
#If True 'False
	Private Sub SaveFrameChannels(framenum As Integer)
'dim svch as Integer 		
'		try
		If Not Me.m_wrseq Then Return 'don't need to write values back to sequence
		If (framenum < 0) or (framenum >= DirectCast(Me.VixenInfo.Seq, Vixen.EventSequence).TotalEventPeriods) Then Return 'eof; no data here
		If Logger.WantTrace Then Logger.LogMsg(String.Format("write {0} channels {1}..{2} to seq frame {3} start channel ofs {4}", MaxChannel - MinChannel + 1, MinChannel, MaxChannel, framenum, Me.m_startch)) ', LogMode.Flush)
		For ch As Integer = MinChannel To MaxChannel
'	svch = ch
			DirectCast(Me.VixenInfo.Seq, Vixen.EventSequence).EventValues(ch - 1 + Me.m_startch - 1, framenum) = Me.FrameBuf(ch - MinChannel) 'frame# is 0-based but it's one behind, channel#s are 1-based
		Next ch
'		Windows.Forms.Application.DoEvents()
'		Catch exc As Exception
'			msgbox(exc.Message & ", startch " & me.m_startch & ", frame " & framenum & ", Min " & minchannel & ", max " & maxchannel & ", ch " & svch)
'		end try
	End Sub
#End If

'kludge: Vixen seems to be ignoring small channel value changes, so make the change more noticeable
'	Private Function TtrkAdjust(framenum As Long) As Long
''		If (framenum And 1) <> 0 Then Return framenum Xor &H800000
'		Return framenum
'	End Function

'	Public Function RelChannel(chofs As Integer, frame As Integer) As Byte
'		Return AbsChannel(Me.VixenInfo.StartChannel + chofs, frame)
'	End Function
'	Public Function AbsChannel(chofs As Integer, frame As Integer) As Byte
	Public Function HostChannel(chofs As Integer, frame As Integer) As Byte 'C# doesn't like optional params, so overload instead
		Return HostChannel(chofs, frame, 0)
	End Function
	Public Function HostChannel(chofs As Integer, frame As Integer, defval As Byte) As Byte
		If (chofs < 1) Or (chofs > Me.VixenInfo.Seq.Channels.Count) Then Return defval 'return hopefully benign value non-existent cells
		If (frame < 0) Or (frame >= DirectCast(Me.VixenInfo.Seq, Vixen.EventSequence).TotalEventPeriods) Then Return defval 'return hopefully benign value non-existent cells
		Return DirectCast(Me.VixenInfo.Seq, Vixen.EventSequence).EventValues(chofs - 1, frame) 'frame# is 0-based, channel#s are 1-based
	End Function

'write value to sequence, read back and verify:
	Private Sub WriteConfig(ByVal name As String, ByVal value As Integer)
		Me.VixenInfo.SetupData.SetInteger(Me.m_setupNode, name, value)
		Dim verify As Integer = Me.VixenInfo.SetupData.GetInteger(Me.m_setupNode, name, value + 1)
		If verify <> value Then Throw New Exception(String.Format("setupData.SetInteger failed: got {0}, expected {1}", verify, value)) 'paranoid: verify file I/O
	End Sub


'write value to sequence, read back and verify:
	Private Sub WriteConfig(ByVal name As String, ByVal value As Boolean)
		Me.VixenInfo.SetupData.SetBoolean(Me.m_setupNode, name, value)
		Dim verify As Boolean = Me.VixenInfo.SetupData.GetBoolean(Me.m_setupNode, name, not value)
		If verify <> value Then Throw New Exception(String.Format("setupData.SetBoolean failed: got {0}, expected {1}", verify, value)) 'paranoid: verify file I/O
	End Sub


'write value to sequence, read back and verify:
	Private Sub WriteConfig(ByVal name As String, ByVal value As String)
		Me.VixenInfo.SetupData.SetString(Me.m_setupNode, name, value)
		Dim verify As String = Me.VixenInfo.SetupData.GetString(Me.m_setupNode, name, value & "x")
		If verify <> value Then Throw New Exception(String.Format("setupData.SetString failed: got {0}, expected {1}", verify, value)) 'paranoid: verify file I/O
	End Sub


'avoid divide-by-0 errors:
	Private Function SafeDiv(ByVal num As Long, ByVal den As Long) As Long
		If den = 0 Then Return 0 'avoid divide by 0 error if no stats collected yet
		Return CLng(num/den) 'rounded divide; use num\den to avoid rounding
	End Function
#End Region
End Class


'frame tags/notes/cues:
'track = individual RGB prop, Papagayo "voice", channel group, etc.
Public Class FrameTag
	Public TagNode As XmlNode
	Public StartFrame As Integer, EndFrame As Integer
	Public Shared TagCount As Integer = 0
	Public Shared ReadOnly Property TrackCount As Integer
		Get
			Return m_byframe.Count
		End Get
	End Property
'	Public Sub PropagateTags(src As FxGenPluginBase)
'		TagCount = src.TagCount
'		m_byframe = src.m_byframe
'		m_byvalue = src.m_byvalue
'	End Sub

	Public ReadOnly Property Text As String
		Get
			Return Me.TagNode.InnerText
		End Get
	End Property
	Public Sub New(tagnode As XmlNode, vixinfo As VixenInfo) 'maxframe As Integer)
		Me.TagNode = tagnode 'allow custom attrs to be accessed later
'		m_frtags.Add(Me)
		Dim trk As String = SafeValue(tagnode.Attributes("track"), "none") 'default track name if missing
		Dim trklist As FixedupSortedList(Of Integer, FrameTag) = Nothing 'init to Nothing to avoid compiler warning
		If Not m_byframe.TryGetValue(trk, trklist) Then
			trklist = New FixedupSortedList(Of Integer, FrameTag)
			m_byframe.Add(trk, trklist)
		End If
		If tagnode.Attributes("value") IsNot Nothing Then 'index by value (must be unique)
			Dim val As Integer = Integer.Parse("0" & tagnode.Attributes("value").Value) 'default value to 0 if missing
			Dim key As String = String.Format(ValueKeyFmt, trk, val)
			Dim other As FrameTag = Nothing 'init to Nothing to avoid compiler warning
			If m_byvalue.TryGetValue(key, other) Then Throw New Exception(String.Format("Duplicate frame tag track '{0}', value {1}: '{2}' vs. '{3}'", trk, val, tagnode.InnerText, other.TagNode.InnerText))
			m_byvalue.Add(key, Me)
		Else
			If (trklist.Count > 0) And (tagnode.Attributes("start") Is Nothing) And (tagnode.Attributes("end") Is Nothing) Then Throw New Exception(String.Format("Frame tag must have a value or a start/end frame: '{0}'", tagnode.InnerText))
		End If
		Me.StartFrame = vixinfo.Time2Frame_seq(SafeValue(tagnode.Attributes("start"), "0"), tagnode.InnerText) 'default to start of sequence if missing
		Me.EndFrame = vixinfo.Time2Frame_seq(SafeValue(tagnode.Attributes("end"), DirectCast(vixinfo.Seq, Vixen.EventSequence).TotalEventPeriods.ToString()), tagnode.InnerText) 'default to end of sequence if missing
		If trklist.Count > 0 Then 'check for duplicate range
			Dim inx As Integer = Array.BinarySearch(Of Integer)(trklist.KeyArray, Me.StartFrame) 'Xor -1
			If inx >= 0 Then
				If (trklist(inx).StartFrame > 0) Or (trklist(inx).EndFrame < DirectCast(vixinfo.Seq, Vixen.EventSequence).TotalEventPeriods) Then
					Throw New Exception(String.Format("Duplicate frame tag range for track '{0}', start frame {1}: '{2} vs. '{3}'", trk, Me.StartFrame, Me.TagNode.InnerText, trklist(inx).TagNode.InnerText))
				Else
					trklist.RemoveAt(inx) 'bump entire range to make room for narrower range
				End If
			End If
		End If
		trklist.Add(Me.StartFrame, Me)
		If Logger.WantTrace Then Logger.LogMsg(String.Format("load frametag[{0}]: track '{1}', start {2}, end {3}, val {4}, text '{5}'", TagCount, trk, Me.StartFrame, Me.EndFrame, SafeValue(tagnode.Attributes("value"), String.Empty), tagnode.InnerText))
		TagCount += 1
	End Sub

'	Private Shared m_frtags As New List(Of FrameTag)
	Public Shared Sub Clear()
'		m_frtags.Clear()
		TagCount = 0
		m_byframe.Clear()
		m_byvalue.Clear()
	End Sub

	Private Shared m_byframe As New Dictionary(Of String, FixedupSortedList(Of Integer, FrameTag))
	Public Shared Function ByFrame(track As String, frame As Integer) As FrameTag
		Dim trklist As FixedupSortedList(Of Integer, FrameTag) = Nothing 'set to Nothing to avoid compiler warning
		If Not m_byframe.TryGetValue(track, trklist) Then Return Nothing
		If trklist.Count < 1 Then Return Nothing
		Dim inx As Integer = Array.BinarySearch(trklist.KeyArray, frame)
'		Dim keylist As New StringBuilder()
'		For i As Integer = 1 To trklist.Keys.Count
'			keylist.AppendFormat(", {0}@{1}", trklist.Keys(i - 1), trklist(trklist.Keys(i - 1)).StartFrame)
'		Next i
'		Logger.LogMsg(String.Format("byframe(trk '{0}', frame {1}): srch picked {2}/{3} from list {4}", track, frame, inx, trklist.Count, keylist.ToString()))
		If inx < 0 Then 'use entry prior to requested frame# if no exact match; list is non-empty
			inx = Math.Min(-inx - 1, trklist.Count - 1) 'don't go past last entry in list
			Do
				If trklist(trklist.keys(inx)).StartFrame <= frame Then Exit Do 'requested frame is before this one
				inx -= 1
				If inx < 0 Then Return Nothing 'no tag is active for requested frame#
			Loop
			If trklist(trklist.keys(inx)).EndFrame < frame Then Return Nothing 'no tag is active for requested frame#
		End If
		Return trklist(trklist.keys(inx)) 'tag is active for requested frame
	End Function

	Private Const ValueKeyFmt As String = "{0}.{1:D5}"
	Private Shared m_byvalue As New Dictionary(Of String, FrameTag)
	Public Shared Function ByValue(track As String, value As Integer) As FrameTag
		ByValue = Nothing
		If m_byvalue.TryGetValue(String.Format(ValueKeyFmt, track, value), ByValue) Then Exit Function
		Return Nothing
	End Function
	Public Shared Function ByValueAsText(track As String, value As Integer) As String
		Dim frtag As FrameTag = ByValue(track, value)
		If frtag Is Nothing Then Return String.Empty
		Return frtag.TagNode.InnerText
	End Function

	Public Shared Function SafeValue(obj As XmlAttribute) As String
		Return SafeValue(obj, String.Empty)
	End Function
	Public Shared Function SafeValue(obj As XmlAttribute, def As String) As String
		If obj Is Nothing Then Return def 'String.Empty
		If Not String.IsNullOrEmpty(obj.InnerText) Then Return obj.InnerText
		If Not String.IsNullOrEmpty(obj.Value) Then Return obj.Value
		Return def
	End Function
End Class
'End Class 'FxGen_base


'wrapper for Vixen run-time info:
Public Class VixenInfo
'info direct from Vixen:
	Public SetupData As Vixen.SetupData = Nothing
	Public Seq As Vixen.IExecutable = Nothing 'Vixen sequence or profile object
'	Public Seq As Vixen.EventSequence = Nothing 'Vixen sequence object
'other info from Vixen environment or sequence file:
	Public Channels As New List(Of Channel) 'Vixen channels
	Public StartChannel as Integer = 0 'my start channel# (0-based)
	Public EndChannel As Integer = 0 'my end channel# (0-based)
	Public TimingTrack As Integer = 0 'start offset of timing track (0-based)
	Public PortsUsed As New Dictionary(Of String, String)
'state info:
	Public CurrentFrame As Integer = 0 'current frame# within sequence
	public FrameTags as String = String.Empty 'tags/macro functions for current frame
'overall performance stats:
	Public total_evts As Long = 0 'total #events processed
	Public total_inlen As Long = 0 'real channels processed
	Public total_outlen As Long = 0 'virtual channels processed
	Public total_nullout As Long = 0 'total #channel output bytes that were full off
	Public total_macros As Long = 0 'total #macro functions/effects processed

'check if a serial port is virtual:
'temp kludge: port is assumed virtual if it is >= 10 and there is a corresponding port >= 20 (com0com pair)
	Public Function IsVirtual(name As String) As String
'		Return Integer.Parse("0" & name.Replace("COM", String.Empty)) >= 20 'kludge: assume anything >= 20 is virtual; >= 10 might are typically USB-to-serial ports
		If Not name.StartsWith("COM") Then Return String.Empty
		Dim othername as String = String.Format("COM{0}", Integer.Parse(name.Substring(3)) + 10)
		Dim other As New System.IO.Ports.SerialPort(othername) 'NOTE: returns another port if requested port not found~
		If (other IsNot Nothing) AndAlso (other.PortName = othername) Then Return other.PortName
		Return String.Empty
	End Function

'convert a path name that is relative to sequence to absolute path:
	Public Function Rel2Abs(path As String) As String
		If path.StartsWith("..\") Then Return System.IO.Path.Combine(System.IO.Path.GetDirectoryName(System.IO.Path.GetDirectoryName(Me.Seq.FileName)), path.Substring(3)) 'use parent folder
		If path.StartsWith(".\") Then Return System.IO.Path.Combine(System.IO.Path.GetDirectoryName(Me.Seq.FileName), path.Substring(2)) 'use current folder
		Return path
	End Function

	Public Function Abs2Rel(path As String) As String
		Dim lcpath As String = path.ToLower()
		Dim parent As String = System.IO.Path.GetDirectoryName(Me.Seq.FileName.ToLower())
		If lcpath.StartsWith(parent) Then Return "." & path.Substring(parent.Length)
		parent = System.IO.Path.GetDirectoryName(System.IO.Path.GetDirectoryName(Me.Seq.FileName.ToLower()))
		If lcpath.StartsWith(parent) Then Return ".." & path.Substring(parent.Length)
		Return path
	End Function

'convert elapsed timestamp to frame#:
'expected format: integer or #:##.### with optional junk in parentheses after
'	Private Shared tsfmt As New RegularExpressions.Regex("^(\d{0,2})?:?(\d{0,2})\.?(d{0,3})?(?:\s*\(\d+\)\s*)$|^(\d+)$")
	Private Shared tsfmt As New RegularExpressions.Regex("^(\d+)$|^(\d{0,2})?:?(\d{0,2})\.?(\d{0,3})?\s*(\(\d+\))?\s*$")
	Public Function Time2Frame_seq(timestamp As String) As Integer
		Return time2frame_seq(timestamp, String.Empty)
	End Function
	Public Function Time2Frame_seq(timestamp As String, desc As String) As Integer
		Try
10:			Dim parts As Match = tsfmt.Match(timestamp)
11:			If (Not parts.Success) OrElse (parts.Groups.Count <> 5+1) OrElse (parts.Groups(1).Length + parts.Groups(2).Length + parts.Groups(3).Length + parts.Groups(4).Length = 0) Then Throw New Exception(String.Format("Invalid timestamp or frame#: '{0}' ({1} {2}", timestamp, parts.Success, parts.Groups.Count))
12:			If parts.groups(1).Length <> 0 Then 'bare frame#
13:				Return Convert.ToInt32(parts.Groups(1).Value)
			End If
14:			Dim framenum As Integer = 0
15:			If parts.groups(2).Length <> 0 Then framenum += 60000 * Convert.ToInt32(parts.Groups(2).Value)
16:			If parts.groups(3).Length <> 0 Then framenum += 1000 * Convert.ToInt32(parts.Groups(3).Value)
17:			If parts.groups(4).Length <> 0 Then framenum += Convert.ToInt32(parts.Groups(4).Value)
18:			Return framenum \ DirectCast(Me.Seq, Vixen.EventSequence).EventPeriod 'convert relative time to frame#
		Catch exc As Exception
			If Not String.IsNullOrEmpty(desc) Then desc = String.Format(" '{0}'", desc)
			Logger.ReportError(String.Format("Time2Frame '{0}'{1}", timestamp, desc), exc)
			Return 0
		End Try
	End Function
End Class


'attach a color to a text string:
Public Class ColoredString
'	Inherits String, Color 'can't inh from these

	Public color As Color
	Public text As String
	Private defcolor As Color = color.White 'FromArgb(255, &HC0, &HF0, &HFF) 'light blue

	Public Sub New(str As String, c As Color)
		Me.text = str
		Me.color = c
	End Sub
	Public Sub New(str As String, r As Byte, g As Byte, b As Byte)
		Me.text = str
		Me.color = color.FromArgb(255, r, g, b)
	End Sub
	Public Sub New(str As String)
		Me.text = str
		Me.color = defcolor 'color.White
	End Sub
	Public Sub New(c As Color)
		Me.text = String.Empty
		Me.color = c
	End Sub
	Public Sub New(r As Byte, g As Byte, b As Byte)
		Me.text = String.Empty
		Me.color = color.FromArgb(255, r, g, b)
	End Sub
	Public Sub New
		Me.text = String.Empty
		Me.color = defcolor 'Color.White
	End Sub
End Class

'fixed up SortedList to work with BinarySearch:
Public Class FixedupSortedList(Of TKey, TValue)
	Inherits SortedList(Of TKey, TValue)

	Private m_keys As TKey()
	Private m_dirty As Boolean = True
	
	Public Overloads Sub Add(ByVal key As TKey, ByVal value As TValue)
		MyBase.Add(key, value)
		Me.m_dirty = True 'lazy copy
	End Sub
	
	Public Function KeyArray As TKey()
		If Me.m_dirty Then 'regenerate key array (lazy copy)
			Me.m_keys = New TKey(MyBase.Count){}
			For Each key As TKey In MyBase.Keys
				Me.m_keys(MyBase.Keys.IndexOf(key)) = key
			Next key
			Me.m_dirty = False
		End If
		Return Me.m_keys
	End Function
End Class

End Namespace 'FxGen
