'
'Vixen Effects Generator plug-in, Setup window
'
'Copyright (c) 2011 Don Julien.  djulien@thejuliens.net
'
'see VixFxGen.vb for additional info

Option Strict On
Option Explicit On
Option Compare Text

Imports System.Windows.Forms
Imports System.Text


Public Partial Class VixFxGenSetup
'	Private components As IContainer

#Region "Instantiation"
'	Private m_log As FxGen.FxLogger = Nothing 'shared with parent
	Private m_chmap As FxGen_base.FxGen.FxGenPluginBase 'provides channel mapping/ordering options
    Public Sub New(chmapper As FxGen_base.FxGen.FxGenPluginBase) 'log As FxGen.FxLogger)
		Me.InitializeComponent() 'required for Windows Forms designer support.
		Me.m_chmap = chmapper
		Dim pixtypes As New SortedList(Of String, String)
		For Each pixtype As String In Me.m_chmap.PixelTypes
			pixtypes.Add(pixtype, pixtype)
		Next pixtype
		For i As Integer = 1 To pixtypes.Count
			Me.pixdepthComboBox.Items.Add(pixtypes.Values(i - 1)) 'these never change, so just load them once
		Next i
		Me.SetState(Me.dummyComboBox, "choose Virt Profile first")
		Me.SetState(Me.pixdepthComboBox, "choose Plug-in first")
		Me.SetState(Me.canvasComboBox, "choose Pixel Type first")
		Me.SetState(Me.chorderComboBox, "choose Pixel Type and Canvas Size first")
		Me.IsDirty = False
'		Me.m_log = log
	End Sub
#End Region

#Region "Properties"
'these are just wrappers around editable controls; they represent configuarion parameters
'wrappers are used do that dirty state can be tracked
	Public Property PauseOnError As Boolean
		Get
			Return Me.errpauseCheckBox.Checked
		End Get
		Set(ByVal newval As Boolean)
			Me.errpauseCheckBox.Checked = newval
		End Set
	End Property

	Public Property AudioControl As Boolean
		Get
			Return Me.audioCheckBox.Checked
		End Get
		Set(ByVal newval As Boolean)
			Me.audioCheckBox.Checked = newval
		End Set
	End Property

	Public Property DataLogging As Boolean
		Get
			Return Me.dataLoggingCheckBox.Checked
		End Get
		Set(ByVal newval As Boolean)
			Me.dataLoggingCheckBox.Checked = newval
		End Set
	End Property

	Public Property ReadSeq As Boolean
		Get
			Return Me.rdseqCheckBox.Checked
		End Get
		Set(ByVal newval As Boolean)
			Me.rdseqCheckBox.Checked = newval
		End Set
	End Property

	Public Property WriteSeq As Boolean
		Get
			Return Me.wrseqCheckBox.Checked
		End Get
		Set(ByVal newval As Boolean)
			Me.wrseqCheckBox.Checked = newval
		End Set
	End Property

	Public Property Chofs As Integer
		Get
			Return CInt("0" & Me.ChofsTextBox.Text)
		End Get
		Set(ByVal newval As Integer)
			Me.ChofsTextBox.Text = newval.ToString()
		End Set
	End Property

	Private m_maxchofs As Integer
	Public Property MaxChofs As Integer
		Get
			Return Me.m_maxchofs
		End Get
		Set(ByVal newval As Integer)
			If newval > 0 Then Me.m_maxChofs = newval
		End Set
	End Property

	Public Property LogFile As String
		Get
			Return Me.logFileTextBox.Text
		End Get
		Set(ByVal newval As String)
			Me.logFileTextBox.Text = newval
			Me.file_changed(Me.logFileTextBox, Nothing)
		End Set
	End Property

	Public Property VirtProfile As String
		Get
			Return Me.virtualFileTextBox.Text
		End Get
		Set(ByVal newval As String)
			Me.virtualFileTextBox.Text = newval
			Me.file_changed(Me.virtualFileTextBox, Nothing)
		End Set
	End Property

'	Private m_fxfile As EffectsFileInfo = Nothing
'	Public Property EffectsFile As EffectsFileInfo
'		Get
'			Return Me.m_fxfile
'		End Get
'		Set(ByVal newval As EffectsFileInfo)
'			Me.m_fxfile = newval
'			Me.effectsFileTextBox.Text = newval.Path
'			Me.effectsFileTextBox_changed(Me.effectsFileTextBox, Nothing)
'		End Set
'	End Property

	Private m_PixelTypes As New Dictionary(Of Integer, String)
	Public Property PixelType(inx As Integer) As String
		Get
			If Me.m_PixelTypes.ContainsKey(inx) Then Return Me.m_PixelTypes(inx)
			Return String.Empty
		End Get
		Set(ByVal newval As String)
			Dim newinx As Integer
			Me.pluginsCheckedListBox.SelectedIndex = inx
			If String.IsNullOrEmpty(newval) Then
				newinx = -1
			ElseIf Me.pixdepthComboBox.Items.Contains(newval) Then
				newinx = Me.pixdepthComboBox.Items.IndexOf(newval)
			Else
				Throw New Exception(String.Format("Unknown pixel type: '{0}', choices are: {1}", newval, Choices(Me.pixdepthComboBox.Items, ", ")))
			End If
			Me.pixdepthComboBox.SelectedIndex = newinx
'			Me.pixdepthComboBox_changed(Me.pixdepthComboBox, Nothing)
		End Set
	End Property

	Private m_CanvasSizes As New Dictionary(Of Integer, String)
	Private Shared NoSize As String = FxGen_base.FxGen.FxGenPluginBase.Size2String(New System.Drawing.Size(0, 0))
	Public Property CanvasSize(inx As Integer) As String
		Get
			If Me.m_CanvasSizes.ContainsKey(inx) Then Return Me.m_CanvasSizes(inx)
			Return String.Empty
		End Get
		Set(ByVal newval As String)
			Dim newinx As Integer
			Me.pluginsCheckedListBox.SelectedIndex = inx
			If String.IsNullOrEmpty(newval) OrElse (newval = NoSize) Then
				newinx = -1
			ElseIf Me.canvasComboBox.Items.Contains(newval) Then
				newinx = Me.canvasComboBox.Items.IndexOf(newval)
			Else
				Throw New Exception(String.Format("Unknown canvas geometry: '{0}', choices are: {1}", newval, Choices(Me.canvasComboBox.Items, ",")))
			End If
			Me.canvasComboBox.SelectedIndex = newinx
'			Me.canvasComboBox_changed(Me.canvasComboBox, Nothing)
		End Set
	End Property

	Private m_ChannelOrders As New Dictionary(Of Integer, String)
	Public Property ChannelOrder(inx As Integer) As String
		Get
			If Me.m_ChannelOrders.ContainsKey(inx) Then Return Me.m_ChannelOrders(inx)
			Return String.Empty
		End Get
		Set(ByVal newval As String)
			Dim newinx As Integer
			Me.pluginsCheckedListBox.SelectedIndex = inx
			If String.IsNullOrEmpty(newval) Then
				newinx = -1
			ElseIf Me.chorderComboBox.Items.Contains(newval) Then
				newinx = Me.chorderComboBox.Items.IndexOf(newval)
			Else
				Throw New Exception(String.Format("Unknown channel order: '{0}', choices are: {1}", newval, Choices(Me.chorderComboBox.Items, ", ")))
			End If
			Me.chorderComboBox.SelectedIndex = newinx
'			Me.chorderComboBox_changed(Me.chorderComboBox, Nothing)
'			Me.pluginsCheckedListBox.SelectedIndex = inx
		End Set
	End Property

	Public Property PluginNames(inx As Integer) As String 'array of names
		Get
			If (inx < 0) Or (inx >= Me.pluginsCheckedListBox.Items.Count) Then Return String.Empty
			Return Me.pluginsCheckedListBox.Items(inx).ToString()
		End Get
		Set(ByVal newval As String)
			If inx = 0 Then Me.pluginsCheckedListBox.Items.Clear() 'probably refreshing list; clear out old stuff
			While inx >= Me.pluginsCheckedListBox.Items.Count 'pad items out to desired index
				Me.pluginsCheckedListBox.Items.Add("??", False)
			End While
'			Me.m_PixelTypes.Capacity = Math.Max(Me.m_PixelTypes.Capacity, Me.pluginsCheckedListBox.Items.Count)
'			While Me.m_PixelTypes.Count <= inx
'				Me.m_PixelTypes.Add(-1)
'			End While
'			Me.m_CanvasSizes.Capacity = Math.Max(Me.m_CanvasSizes.Capacity, Me.pluginsCheckedListBox.Items.Count)
'			While Me.m_CanvasSizes.Count <= inx
'				Me.m_CanvasSizes.Add(-1)
'			End While
'			Me.m_ChannelOrders.Capacity = Math.Max(Me.m_ChannelOrders.Capacity, Me.pluginsCheckedListBox.Items.Count)
'			While Me.m_ChannelOrders.Count <= inx
'				Me.m_ChannelOrders.Add(-1)
'			End While
'			While Me.m_numchannels.Count <= inx
'				Me.m_numchannels.Add(0)
'			End While
'			If Me.pluginsCheckedListBox.SelectedIndex < 0 Then Me.pluginsCheckedListBox.SelectedIndex = inx
			Me.pluginsCheckedListBox.SelectedIndex = inx
			Me.pluginsCheckedListBox.Items(inx) = newval
		End Set
	End Property

	Public Property PluginSelected(inx As Integer) As Boolean 'array of boolean
		Get
			If (inx < 0) Or (inx >= Me.pluginsCheckedListBox.Items.Count) Then Return False
			Return Me.pluginsCheckedListBox.GetItemChecked(inx)
		End Get
		Set(ByVal newval As Boolean)
			Me.pluginsCheckedListBox.SelectedIndex = inx
			If (inx < 0) Or (inx >= Me.pluginsCheckedListBox.Items.Count) Then Return
			Me.pluginsCheckedListBox.SetItemChecked(inx, newval)
		End Set
	End Property

	Private m_numchannels As New Dictionary(Of Integer, Integer)
	Public Property PluginChannels(inx As Integer) As Integer
		Get
			If Me.m_numchannels.ContainsKey(inx) Then Return Me.m_numchannels(inx)
			Return 0
		End Get
		Set(ByVal newval As Integer)
'			If (inx < 0) Or (inx >= Me.pluginsCheckedListBox.Items.Count) Then Return
			If Me.m_numchannels.ContainsKey(inx) Then
				Me.m_numchannels(inx) = newval
			Else
				Me.m_numchannels.Add(inx, newval)
			End If
			Me.pluginsCheckedListBox.SelectedIndex = inx
		End Set
	End Property

	Private m_plugtags As New Dictionary(Of Integer, String)
	Public Property PluginTag(inx As Integer) As String
		Get
			If Me.m_plugtags.ContainsKey(inx) Then Return Me.m_plugtags(inx)
			Return String.Empty
		End Get
		Set(ByVal newval As String)
'			If (inx < 0) Or (inx >= Me.pluginsCheckedListBox.Items.Count) Then Return
			If Me.m_plugtags.ContainsKey(inx) Then
				Me.m_plugtags(inx) = newval
			Else
				Me.m_plugtags.Add(inx, newval)
			End If
			Me.pluginsCheckedListBox.SelectedIndex = inx
		End Set
	End Property

'	Private oktoclose As Boolean = True 'KLUDGE: there doesn't seem to be a ButtonClickEventArgs, so custom logic is needed in order to abort a button
'	Private m_dirty As DialogResult = NoChanges
	'states for dirty flag (actually just answers to the exit prompt):
'	Private Const NoChanges As DialogResult = DialogResult.None 'no changes
'	Private Const ApplyChanges As DialogResult = DialogResult.Yes 'exit and keep changes
'	Private Const KeepChanges As DialogResult = DialogResult.Cancel 'don't exit
'	Private Const DiscardChanges As DialogResult = DialogResult.No 'exit and don't save
'	Private Const UnconfirmedChanges As DialogResult = DialogResult.None 'prompt user whether to save
'	Private Const ApplyChanges As DialogResult = DialogResult.Yes 'exit and save
'	Private Const KeepChanges As DialogResult = DialogResult.Cancel 'don't exit
'	Private Const NoChanges As DialogResult = DialogResult.No 'exit and don't save
	'set dirty flag and update dependent menu items:
	Private m_TrackChanges As Boolean = False
	Public Property IsDirty As Boolean
		Get
'			Return Me.m_dirty <> NoChanges
			Return Me.buttonOK.Enabled
		End Get
		Set(newval As Boolean)
'			Me.m_dirty = CType(IIf(newval, ApplyChanges, NoChanges), DialogResult)
			If Logger.WantTrace Then Logger.LogMsg(String.Format("enter: dirty = {0}, track {1}, result {2}", newval, Me.m_TrackChanges, Me.DialogResult))
			If Not Me.m_TrackChanges Then Return
'			Me.ButtonOk.DialogResult = DialogResult.No 'KLUDGE: benign value triggered when button re-enabled
			Me.buttonOK.Enabled = newval 'NOTE: this seems to trigger DialogResult
'			Me.ButtonOk.DialogResult = DialogResult.Yes 'restore normal value
'			Me.DialogResult = CType(IIf(newval, Windows.Forms.DialogResult.Cancel, Windows.Forms.DialogResult.None), Windows.Forms.DialogResult) 'unconfirmed; prompt user whether to save
'			logger.LogMsg(String.Format("exit: dirty = {0}, track {1}, result {2}", newval, Me.m_TrackChanges, Me.DialogResult))
		End Set
	End Property
'	Public Property IsDirty As Boolean
'		Get
'			Return Not Me.oktoclose
'		End Get
'		Set(newval As Boolean)
'			Me.oktoclose = Not newval 'dirty
'			Me.buttonOK.Enabled = newval
'		End Set
'	End Property
#End Region

#Region "Event handlers - form mgmt"
	Public Overloads Function ShowDialog As Windows.Forms.DialogResult
		Me.IsDirty = False 'no user-initiated changes yet
		Return MyBase.ShowDialog
	End Function

	Private Sub changes(ByVal sender As Object, ByVal ea As EventArgs) Handles audioCheckBox.CheckStateChanged, dataLoggingCheckBox.CheckStateChanged, errpauseCheckBox.CheckStateChanged, rdseqCheckBox.CheckStateChanged, wrseqCheckBox.CheckStateChanged, chofsTextBox.TextChanged, virtualFileTextBox.TextChanged, logFileTextBox.TextChanged, pluginsCheckedListBox.ItemCheck, pluginTagTextBox.TextChanged ', pixdepthComboBox.SelectedIndexChanged, canvasComboBox.SelectedIndexChanged, chorderComboBox.SelectedIndexChanged
		If Logger.WantTrace Then Logger.LogMsg(String.Format("change {0}", DirectCast(sender, Control).Name))
		Me.IsDirty = True
	End Sub

	Private Sub buttonOK_Click(ByVal sender As Object, ByVal ea As EventArgs) Handles buttonOK.Click
		Try
'		Me.IsDirty = False 'assume it's okay to close until found otherwise
		Dim sb As New System.Text.StringBuilder
		Dim numsel As Integer = Me.pluginsCheckedListBox.CheckedIndices.Count '0
'		For i As Integer = 1 To Me.pluginsCheckedListBox.Items.Count
'			if Me.pluginsCheckedListBox.GetItemChecked(i - 1) then numsel += 1
'		Next i
		If numsel < 1 Then sb.AppendLine("No output plug-ins have been selected.")
		If (Me.Chofs < 1) Or ((Me.Chofs > Me.m_maxchofs) and (Me.m_maxchofs > 0)) Then sb.AppendLine(String.Format("Start channel must be 1..{0}", Me.MaxChofs))
'		, "No Plug-ins?", System.Windows.Forms.MessageBoxButtons.OK, MessageBoxIcon.Exclamation): Me.oktoclose = False
'		If String.IsNullOrEmpty(Me.effectsFileTextBox.Text) Then
'			sb.AppendLine("No effects file has been selected.")
'		ElseIf (Me.m_fxfile.chmapper Is Nothing) Or (Me.m_fxfile.fxmain Is Nothing) Then
'			sb.AppendLine("Effects file is invalid.")
'		Else
'			If String.IsNullOrEmpty(Me.PixelType) Then sb.AppendLine("No Pixel Type has been selected.")
'			If String.IsNullOrEmpty(Me.CanvasSize) Then sb.AppendLine("No Canvas Size has been selected.")
'			If String.IsNullOrEmpty(Me.ChannelOrder) Then sb.AppendLine("No Channel Order has been selected.")
'		End If
		If sb.Length > 0 Then 'can't save due to errors
			System.Windows.Forms.MessageBox.Show(sb.ToString(), "Invalid Options", System.Windows.Forms.MessageBoxButtons.OK, System.Windows.Forms.MessageBoxIcon.Exclamation)
			Me.DialogResult = Windows.Forms.DialogResult.No 'not okay to close yet
'			Me.m_dirty = ConfirmChanges
		Else
'			Me.DialogResult = Windows.Forms.DialogResult.Yes
'			Me.m_dirty = KeepChanges
'			Me.Hide()
		End If
'		msgbox("TODO: allow virt channels to be set via Setup dlg. for now, just edit the file")
'		Me.UpdateSetup(False)
'		Me.Dirty(False)
		Catch exc As Exception
			Logger.ReportError("ButtonOK", exc)
		End Try
	End Sub

'changes + OK => No confirm:Yes + then save
'changes + Close/Cancel => confirm, then Yes or No
'no changes + Close/Cancel => No confirm:None + quiet exit
'	Private Const NoChanges As DialogResult = DialogResult.None 'no changes
'	Private Const ApplyChanges As DialogResult = DialogResult.Yes 'exit and keep changes
''	Private Const KeepChanges As DialogResult = DialogResult.Cancel 'don't exit
'	Private Const DiscardChanges As DialogResult = DialogResult.No 'exit and don't save

'NOTE: both Cancel and OK come through here
'OK button sets DialogResult.Yes if valid, .No if invalid options
'Cancel button sets DialogResult.Cancel
'NOTE: window doesn't seem to close for DialogResult.None, so use .Ignore instead
	Private Sub form_closing(ByVal sender As Object, ByVal fcea As System.Windows.Forms.FormClosingEventArgs) Handles Me.FormClosing
		Try
		If Logger.WantTrace Then Logger.LogMsg(String.Format("closing, user cancel? {0}, evt cancel? {1}, reason {2} vs. {3}, dirty? {4}", (fcea.CloseReason = CloseReason.UserClosing), fcea.Cancel, fcea.CloseReason, Me.DialogResult, Me.IsDirty))
'		If Me.m_dirty <> DialogResult.None Then Return 'already decided
		If Not Me.IsDirty Then Me.DialogResult = DialogResult.Cancel: Return 'no changes '.Ignore
'		If fcea.CloseReason = CloseReason.UserClosing Then Me.DialogResult = DialogResult.None 'force prompt if user closes window without clicking a button
		Select Case Me.DialogResult
			Case DialogResult.Yes: Return 'user has already confirmed they want changes saved
			Case DialogResult.No: fcea.Cancel = True: Return 'don't close yet
'			Case DialogResult.Cancel: Return 'no changes; confirmation not needed
		End Select
		Me.DialogResult = MessageBox.Show("There are unsaved changes.  Do you want to keep them?", Me.Text, MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question)
		fcea.Cancel = (Me.DialogResult = System.Windows.Forms.DialogResult.Cancel) 'don't close yet
'			Case System.Windows.Forms.DialogResult.Yes
'				If Not Me.UpdateSetup(False) Then e.Cancel = True 'give user a chance to make corrections
'				Me.UpdateSetup(False)
		Catch exc As Exception
			Logger.ReportError("FormClosing", exc)
		End Try
	End Sub
#End Region

#Region "Event handlers - plug-in mgmt"
	Private Sub browse(ByVal sender As Object, ByVal ea As EventArgs) Handles logFileBrowseButton.Click, effectsFileBrowseButton.Click
		Dim islog As Boolean = (DirectCast(sender, Windows.Forms.Control).Name = Me.logFileBrowseButton.Name)
		Dim txtbox As Windows.Forms.TextBox = DirectCast(IIf(islog, Me.logFileTextBox, Me.virtualFileTextBox), Windows.Forms.TextBox)
		Dim fdlg As Windows.Forms.FileDialog = DirectCast(IIf(islog, Me.saveFileDialog, Me.openFileDialog), Windows.Forms.FileDialog)
		If String.IsNullOrEmpty(txtbox.Text) Then
'			fdlg.FileName = txtbox.Text
			fdlg.InitialDirectory = system.Environment.CurrentDirectory 'default to current folder
		Else
			fdlg.FileName = txtbox.Text
			fdlg.InitialDirectory = System.IO.Path.GetDirectoryName(txtbox.Text) 'start in currently configured folder
		End If
		fdlg.DefaultExt = IIf(islog, "log", "pro").ToString()
		fdlg.Filter = String.Format("{0}|All files|*.*", IIf(islog, "Log files (*.log)|*.log", "Vixen profiles (*.pro)|*.pro|Vixen sequences (*.vix)|*.vix"))
		If fdlg.ShowDialog = Windows.Forms.DialogResult.OK Then txtbox.Text = String.Empty: txtbox.Text = fdlg.FileName 'force update
	End Sub

	Private Sub file_changed(ByVal sender As Object, ByVal ea As EventArgs) Handles logFileTextBox.TextChanged, virtualFileTextBox.TextChanged
		Try
		Select Case DirectCast(sender, Control).name
			Case "logFileTextBox":
				If Not String.IsNullOrEmpty(Me.logFileTextBox.Text) Then Me.dataLoggingCheckBox.Checked = True
			Case "virtualFileTextBox":
				Me.SetState(Me.dummyComboBox, IIf(String.IsNullOrEmpty(Me.virtualFileTextBox.Text), "choose Virt Profile first", String.Empty).ToString())
'				Me.pluginsCheckedListBox.Items.Clear()
				Me.m_chmap.LoadVirtProfile(Me.virtualFileTextBox.Text)
				Me.m_chmap.ShowAllPlugs(Me)
				Me.moreplugins(Me.pluginsCheckedListBox, Nothing)
			Case Else:
				Throw New Exception("unhandled item: " & DirectCast(sender, Control).Name) ', "BUG", MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
		End Select
		Catch exc As Exception
			Logger.ReportError("FileChanged", exc)
		End Try
	End Sub

'simulate drop-down behavior for plug-ins list:
	Private Sub showplugins(ByVal sender As Object, ByVal ea As EventArgs) Handles dummyComboBox.DropDown, timer1.Tick
		If ReferenceEquals(sender, Me.timer1) Then 'kludge: can't set focus during DropDown event, so do it a little after
'			Me.pluginsCheckedListBox.Focus()
			Me.pluginsCheckedListBox.Focus()
			Me.pluginsCheckedListBox.Capture = True 'allow mouse to select new entry
			Me.timer1.Enabled = False
			Return
		End If
'update plugin selection from current combobox text:
		If Me.pluginsCheckedListBox.Items.Contains(Me.dummyComboBox.Text) Then
			Me.pluginsCheckedListBox.SelectedIndex = Me.pluginsCheckedListBox.Items.IndexOf(Me.dummyComboBox.Text)
		Else
			Me.pluginsCheckedListBox.SelectedIndex = -1
		End If
'abort dropdown event:
'		Me.dummyComboBox.DroppedDown = False 'no worky (event hasn't occurred yet)
'from http://blog.excastle.com/2004/07/12/aborting-combobox-dropdown-showing-dialog-box/
're-create window handle to prevent dropdown:
		Me.dummyComboBox.IntegralHeight = Not Me.dummyComboBox.IntegralHeight
		Me.dummyComboBox.IntegralHeight = Not Me.dummyComboBox.IntegralHeight
'now show checked listbox instead:
'		Me.dummyComboBox.Visible = False
		Me.pluginsCheckedListBox.Visible = True
		Logger.NoPopup = True 'messes up popup list
'broken		Me.pluginsCheckedListBox.Capture = True 'allow mouse to select new entry immediately
		Me.dummyComboBox.Enabled = False
		Me.timer1.Enabled = True
'		Me.pluginsCheckedListBox.Focus()
	End Sub

	Private Sub hideplugins(ByVal sender As Object, ByVal ea As EventArgs) Handles pluginsCheckedListBox.LostFocus, Me.MouseDown 'dummyComboBox.DropDownClosed
'		Me.dummyComboBox.Visible = True
		Me.dummyComboBox.Enabled = True
		Me.pluginsCheckedListBox.Visible = False
		Logger.NoPopup = False
	End Sub

	Private Sub moreplugins(ByVal sender As Object, ByVal ea As EventArgs) Handles pluginsCheckedListBox.ItemCheck, pluginsCheckedListBox.SelectedIndexChanged
		Me.labelTP.Text = String.Format("{0} of {1} Plug-in{2}:", Me.pluginsCheckedListBox.CheckedItems.Count, Me.pluginsCheckedListBox.Items.Count, IIf(Me.pluginsCheckedListBox.Items.Count <> 1, "s", String.Empty)) '.Replace("0/", "No/")
		Me.dummyComboBox.Items.Clear
		If Me.pluginsCheckedListBox.SelectedIndex = -1 Then Return
		Me.dummyComboBox.Items.Add(Me.pluginsCheckedListBox.SelectedItem.ToString()) 'show latest selection in dummy combo box
		Me.dummyComboBox.SelectedIndex = 0
	End Sub

	Private Sub debug(ByVal sender As Object, ByVal ea As EventArgs) Handles audioCheckBox.CheckedChanged
'		Logger.LogMsg(String.Empty, LogMode.Flush)
	End Sub

	Private Sub selplugin(ByVal sender As Object, ByVal ea As EventArgs) Handles pluginsCheckedListBox.SelectedIndexChanged
'		Dim svdirt As Boolean = Me.IsDirty
		Try
'		Me.SuspendLayout
10:		Dim plugid As Integer = Me.pluginsCheckedListBox.SelectedIndex
11:		If plugid <> -1 Then plugid = FxGen_base.FxGen.FxGenPluginBase.IdFromDesc(Me.PluginNames(plugid))
12:		Me.labelPT.Text = String.Format("Pixel Type {0}:", plugid).Replace(" -1", String.Empty)
13:		Me.labelCS.Text = String.Format("Canvas {0}:", plugid).Replace(" -1", String.Empty)
14:		Me.labelCO.Text = String.Format("Channel Order {0}:", plugid).Replace(" -1", String.Empty)
15:		Me.labelTag.Text = String.Format("Plug-in Tag(s) {0}:", plugid).Replace(" -1", String.Empty)
'		Me.chorderComboBox.Items.Clear
'		Me.canvasComboBox.Items.Clear
'		Me.pixdepthComboBox.Items.Clear
'		Dim pixtypes As String() = Me.m_chmap.PixelTypes
'		For i As Integer = 1 To pixtypes.Length
'			Me.pixdepthComboBox.Items.Add(pixtypes(i - 1))
'		Next i
'reset state to force updates below:
20:		Me.SetState(Me.pixdepthComboBox, IIf(Me.pluginsCheckedListBox.SelectedIndex = -1, "choose Plug-in first", String.Empty).ToString())
21:		If Me.pluginsCheckedListBox.SelectedIndex = -1 Then
'			Me.pixdepthComboBox.SelectedIndex = -1 'force refresh next time
'			Me.canvasComboBox.SelectedIndex = -1 'force refresh next time
'			Me.chorderComboBox.SelectedIndex = -1 'force refresh next time
'			Me.SetState(Me.pixdepthComboBox, "choose Plug-in first")
'			Me.SetState(Me.canvasComboBox, "choose Pixel Type first")
'			Me.SetState(Me.chorderComboBox, "choose Pixel Type and Canvas Size first")
			Return
		End If
'		Me.SetState(Me.pixdepthComboBox)
'13:		Me.SetState(Me.pixdepthComboBox, "choose Plug-in first")
'14:		Me.SetState(Me.canvasComboBox, "choose Pixel Type first")
'15:		Me.SetState(Me.chorderComboBox, "choose Pixel Type and Canvas Size first")
'16:		If Me.pluginsCheckedListBox.SelectedIndex = -1 Then Return
22:		If Logger.WantTrace Then Logger.LogMsg(String.Format("sel plugin: selinx {0}, #ch {1}, pix type {2}, canvas {3}, chorder {4}", Me.pluginsCheckedListBox.SelectedIndex, Me.PluginChannels(Me.pluginsCheckedListBox.SelectedIndex), Me.PixelType(Me.pluginsCheckedListBox.SelectedIndex), Me.CanvasSize(Me.pluginsCheckedListBox.SelectedIndex), Me.ChannelOrder(Me.pluginsCheckedListBox.SelectedIndex))) ', LogMode.Flush)
'17:		Me.m_chmap.NumChannels = Me.PluginChannels(Me.pluginsCheckedListBox.SelectedIndex)
23:'		Me.pixdepthComboBox.SelectedIndex = -1 'force update
'		Dim svdirty As DialogResult = Me.m_dirty
24:		Me.m_TrackChanges = False 'not considered changes
'clear current choices to force reload lists:
25:		Me.pixdepthComboBox.SelectedIndex = -1
26:		Me.canvasComboBox.SelectedIndex = -1
27:		Me.chorderComboBox.SelectedIndex = -1
28:		Me.pixdepthComboBox.SelectedIndex = Me.pixdepthComboBox.Items.IndexOf(Me.PixelType(Me.pluginsCheckedListBox.SelectedIndex))
'		Me.SetState(Me.pixdepthComboBox, String.Empty)
'		Me.canvasComboBox.SelectedIndex = -1 'force update
29:		Me.canvasComboBox.SelectedIndex = Me.canvasComboBox.Items.IndexOf(Me.CanvasSize(Me.pluginsCheckedListBox.SelectedIndex))
'		Me.chorderComboBox.SelectedIndex = -1 'force update
30:		Me.chorderComboBox.SelectedIndex = Me.chorderComboBox.Items.IndexOf(Me.ChannelOrder(Me.pluginsCheckedListBox.SelectedIndex))
		Me.pluginTagTextBox.Text = Me.PluginTag((Me.pluginsCheckedListBox.SelectedIndex))
31:		Me.m_TrackChanges = True 'dirty = svdirt 'no change here
'		Me.ResumeLayout(False)
'		Exit Sub
		Catch exc As Exception
			Logger.ReportError("SelPlugin", exc)
		End Try
'		Me.IsDirty = svdirt 'no data changed
	End Sub

	Private Sub pixdepthComboBox_changed(ByVal sender As Object, ByVal ea As EventArgs) Handles pixdepthComboBox.SelectedIndexChanged
		Try 'On Error GoTo errh
		If Logger.WantTrace Then Logger.LogMsg(String.Format("pix changed: sel inx {0}", Me.pixdepthComboBox.SelectedIndex))
'		If Me.pluginsCheckedListBox.SelectedIndex = -1 Then
'			Me.SetState(Me.pixdepthComboBox, "choose Plug-in first")
'		Else
'			Me.SetState(Me.pixdepthComboBox)
'		End If
		Me.canvasComboBox.SelectedIndex = -1 'force refresh next time
'		Me.SetState(Me.pixdepthComboBox, IIf(Me.pluginsCheckedListBox.SelectedIndex = -1, "choose Plug-in first", String.Empty).ToString())
		Me.SetState(Me.canvasComboBox, IIf((Me.pluginsCheckedListBox.SelectedIndex = -1) Or (Me.pixdepthComboBox.SelectedIndex = -1), "choose Pixel Type first", String.Empty).ToString())
		If Me.pixdepthComboBox.SelectedIndex = -1 Then
'			Me.canvasComboBox.SelectedIndex = -1 'force refresh next time
'			Me.SetState(Me.canvasComboBox, "choose Pixel Type first")
'			Me.SetState(Me.chorderComboBox, "choose Pixel Type first")
			Return
		End If
'		Me.SetState(Me.pixdepthComboBox)
'37:		Me.SetState(Me.canvasComboBox, String.Empty)
		If Me.PixelType(Me.pluginsCheckedListBox.SelectedIndex) <> Me.pixdepthComboBox.SelectedItem.ToString() Then Me.IsDirty = True
29:		Me.m_PixelTypes(Me.pluginsCheckedListBox.SelectedIndex) = Me.pixdepthComboBox.SelectedItem.ToString() 'remember new value
'		Me.m_CanvasSizes.Remove(Me.pluginsCheckedListBox.SelectedIndex) 'no longer valid
'		Me.m_ChannelOrders.Remove(Me.pluginsCheckedListBox.SelectedIndex) 'no longer valid
'		Me.chorderComboBox.Items.Clear
30:		Me.canvasComboBox.Items.Clear
'?		Me.canvasComboBox.SelectedIndex = -1
'		Logger.LogMsg(String.Format("sel pix depth: #ch {0}", Me.m_chmap.NumChannels))
31:		Me.m_chmap.PixelType = Me.pixdepthComboBox.SelectedItem.ToString()
		Me.m_chmap.NumChannels = Me.m_numchannels(Me.pluginsCheckedListBox.SelectedIndex)
'		Dim geoms As System.Drawing.Size() = Me.m_chmap.Geometries
'		For i As Integer = 1 To geoms.Length
'			Me.canvasComboBox.Items.Add(FxGen_base.FxGen.FxGenPluginBase.Size2String(geoms(i - 1))) 'Me.m_chmap.
'		Next i
32:		Dim geoms As New SortedList(Of String, String)
33:		For Each geom As System.Drawing.Size In Me.m_chmap.Geometries 'normalized sort
34:			geoms.Add(String.Format("{0:D5}x{1:D5}", geom.Width, geom.Height), FxGen_base.FxGen.FxGenPluginBase.Size2String(geom))
		Next geom
'		Logger.LogMsg(String.Format("pix changed: val {0} has {1} canvas choices", Me.m_chmap.PixelType, geoms.Count))
35:		For i As Integer = 1 To geoms.Count
36:			Me.canvasComboBox.Items.Add(geoms.Values(i - 1))
		Next i
		If Logger.WantTrace Then Logger.LogMsg(String.Format("pix changed: set canvas {0} inx to {1}, canv choices: {2}", Me.CanvasSize(Me.pluginsCheckedListBox.SelectedIndex), Me.canvasComboBox.Items.IndexOf(Me.CanvasSize(Me.pluginsCheckedListBox.SelectedIndex)), Choices(Me.canvasComboBox.Items)))
'		Dim svdirt As DialogResult = Me.m_dirty
20:		Me.canvasComboBox.SelectedIndex = Me.canvasComboBox.Items.IndexOf(Me.CanvasSize(Me.pluginsCheckedListBox.SelectedIndex))
'		Me.m_dirty = svdirt 'no change here
'		Me.SetState(Me.chorderComboBox, "pick Canvas Size first")
'38:		Me.SetState(Me.chorderComboBox, "choose Canvas Size first")
'		Return
'errh:	Logger.ReportError("PixChange")
		Catch exc As Exception
			Logger.ReportError("PixChange", exc)
		End Try
	End Sub

	Private Sub canvasComboBox_changed(ByVal sender As Object, ByVal ea As EventArgs) Handles canvasComboBox.SelectedIndexChanged
		Try 'On Error GoTo errh
		If Logger.WantTrace Then Logger.LogMsg(String.Format("canv changed: sel inx {0}", Me.canvasComboBox.SelectedIndex))
		Me.chorderComboBox.SelectedIndex = -1 'force refresh next time
'		Me.SetState(Me.canvasComboBox, IIf((Me.pluginsCheckedListBox.SelectedIndex = -1) Or (Me.pixdepthComboBox.SelectedIndex = -1), "choose Pixel Type first", String.Empty).ToString())
		Me.SetState(Me.chorderComboBox, IIf((Me.pluginsCheckedListBox.SelectedIndex = -1) Or (Me.pixdepthComboBox.SelectedIndex = -1) Or (Me.canvasComboBox.SelectedIndex = -1), "choose Canvas Size first", String.Empty).ToString())
		If Me.canvasComboBox.SelectedIndex = -1 Then
'			Me.SetState(Me.canvasComboBox, "choose Pixel Type first")
'			Me.chorderComboBox.SelectedIndex = -1 'force refresh next time
'			Me.SetState(Me.chorderComboBox, "choose Canvas Size first")
			Return
		End If
'		Me.SetState(Me.canvasComboBox)
		If Me.CanvasSize(Me.pluginsCheckedListBox.SelectedIndex) <> Me.canvasComboBox.SelectedItem.ToString() Then Me.IsDirty = True
39:		Me.m_CanvasSizes(Me.pluginsCheckedListBox.SelectedIndex) = Me.canvasComboBox.SelectedItem.ToString() 'remember new value
'		Me.m_ChannelOrders.Remove(Me.pluginsCheckedListBox.SelectedIndex) 'no longer valid
40:		Me.chorderComboBox.Items.Clear
41:		Me.m_chmap.Geometry = FxGen_base.FxGen.FxGenPluginBase.String2Size(Me.canvasComboBox.SelectedItem.ToString()) 'Me.m_chmap.
'		Me.chorderComboBox.SelectedIndex = -1
'		Dim chords As String() = Me.m_chmap.ChannelOrders
42:		Dim chords As New SortedList(Of String, String)
43:		For Each chord As String In Me.m_chmap.ChannelOrders
44:			chords.Add(chord, chord)
		Next chord
'		Logger.LogMsg(String.Format("canv changed: val {0} has {1} canvas choices", Me.m_chmap.Geometry.ToString(), chords.Count))
45:		For i As Integer = 1 To chords.Count
46:			Me.chorderComboBox.Items.Add(chords.Values(i - 1))
		Next i
		If Logger.WantTrace Then Logger.LogMsg(String.Format("canv changed: set chord {0} inx to {1}, chord choices: {2}", Me.ChannelOrder(Me.pluginsCheckedListBox.SelectedIndex), Me.chorderComboBox.Items.IndexOf(Me.ChannelOrder(Me.pluginsCheckedListBox.SelectedIndex)), Choices(Me.chorderComboBox.Items)))
'		Dim svdirt As DialogResult = Me.m_dirty
20:		Me.chorderComboBox.SelectedIndex = Me.chorderComboBox.Items.IndexOf(Me.ChannelOrder(Me.pluginsCheckedListBox.SelectedIndex))
'		Me.m_dirty = svdirt 'no change here
'		If (Me.m_fxfile.chmapper IsNot Nothing) And (Me.canvasComboBox.Items.Count > 1) Then
'			Me.m_fxfile.chmapper.Geometry = FxGen.FxUtils.String2Size(Me.canvasComboBox.SelectedItem.ToString())
'			Dim chorders As String() = Me.m_fxfile.chmapper.ChannelOrders()
'			If (Not String.IsNullOrEmpty(Me.canvasComboBox.SelectedItem.ToString())) And (chorders.Length < 1) Then Throw New Exception(String.Format("No channel orderss returned for pixel type {0} and canvas {1} by {2} channel mapper", Me.PixelType, Me.chorderComboBox.SelectedItem, Me.m_fxfile.chmapper.GetType.Name))
'			For i As Integer = 1 To chorders.Length
'				Me.chorderComboBox.Items.Add(chorders(i - 1))
'			Next i
'			Me.SetState(Me.chorderComboBox)
'		End If
'		Me.SetState(Me.chorderComboBox, "pick Pixel Type and Canvas Size first")
'47:		Me.SetState(Me.chorderComboBox, String.Empty)
		Catch exc As Exception
			Logger.ReportError("CanvasChange", exc)
		End Try
	End Sub

	Private Sub chorderComboBox_changed(ByVal sender As Object, ByVal ea As EventArgs) Handles chorderComboBox.SelectedIndexChanged
		Try 'On Error GoTo errh
		If Logger.WantTrace Then Logger.LogMsg(String.Format("chord changed: sel inx {0}", Me.chorderComboBox.SelectedIndex))
'		Me.SetState(Me.chorderComboBox, IIf((Me.pluginsCheckedListBox.SelectedIndex = -1) Or (Me.pixdepthComboBox.SelectedIndex = -1) Or (Me.canvasComboBox.SelectedIndex = -1), "choose Canvas Size first", String.Empty).ToString())
		If Me.chorderComboBox.SelectedIndex = -1 Then
'			Me.SetState(Me.canvasComboBox, "choose Canvas Size first")
'			Me.chorderComboBox.SelectedIndex = -1 'force refresh next time
			Return
		End If
'		Me.SetState(Me.chorderComboBox)
		If Me.ChannelOrder(Me.pluginsCheckedListBox.SelectedIndex) <> Me.chorderComboBox.SelectedItem.ToString() Then Me.IsDirty = True
		Me.m_ChannelOrders(Me.pluginsCheckedListBox.SelectedIndex) = Me.chorderComboBox.SelectedItem.ToString() 'remember new value
		Catch exc As Exception
			Logger.ReportError("ChorderChange", exc)
		End Try
	End Sub

	Private Sub pluginTagTextBox_changed(ByVal sender As Object, ByVal ea As EventArgs) Handles pluginTagTextBox.TextChanged
		Try 'On Error GoTo errh
		If Logger.WantTrace Then Logger.LogMsg(String.Format("tag changed: {0}", Me.pluginTagTextBox.Text))
		If Me.PluginTag(Me.pluginsCheckedListBox.SelectedIndex) <> Me.pluginTagTextBox.Text Then Me.IsDirty = True
		Me.m_plugtags(Me.pluginsCheckedListBox.SelectedIndex) = Me.pluginTagTextBox.Text 'remember new value
		Catch exc As Exception
			Logger.ReportError("PluginTagChange", exc)
		End Try
	End Sub
#End Region

#Region "Helper functions"
	Private Sub SetState(ctl As Windows.Forms.ComboBox)
		Me.SetState(ctl, String.Empty)
	End Sub
	Private Sub SetState(ctl As Windows.Forms.ComboBox, msg As String)
		Dim lbl As Control = Me.Controls.Find(ctl.Tag.ToString(), True) (0)
		If String.IsNullOrEmpty(msg) Then 'enable
			ctl.Enabled = True
			If lbl IsNot Nothing Then lbl.Enabled = True
		Else
'			ctl.Items.Add(String.Format("({0})", msg)
'			ctl.SelectedIndex = 0
			ctl.Enabled = False
			If lbl IsNot Nothing Then lbl.Enabled = False
'			ctl.Items.Clear() 'force reload next time
		End If
'		If String.IsNullOrEmpty(ctl.Tag) Then Return
	End Sub

	Private Function Choices(items As ComboBox.ObjectCollection) As String
		Return Choices(items, ", ")
	End Function
	Private Function Choices(items As ComboBox.ObjectCollection, delim As String) As String
		If items.Count = 0 Then Return "(none)"
		Dim sb As New StringBuilder()
		sb.AppendFormat("{0} ", items.Count)
		For i As Integer = 1 To items.Count
			If i > 1 Then sb.Append(delim)
			sb.Append(items(i - 1).ToString())
		Next i
		Return sb.ToString()
	End Function
#End Region
End Class