﻿'
' Created by SharpDevelop.
' User: dj
' Date: 8/17/2011
' Time: 11:40 AM
' 
' To change this template use Tools | Options | Coding | Edit Standard Headers.

Option Strict On
Option Explicit On
Option Compare Text


Imports System
'Imports System.Collections.Generic
'Imports System.IO.File
Imports System.Text
Imports System.Text.RegularExpressions
'Imports System.Runtime.InteropServices
Imports System.Windows.Forms
Imports System.Threading
Imports System.Xml
Imports System.Reflection
Imports System.Drawing

Imports Vixen 'this will be 2.1 or 2.5, depending on the project references

Namespace FxGen

#If False
Public Class FxLogger
	Public ErrPause As Boolean = False 'stop on error, or just log it and keep going
	Public WantTrace As Boolean = False 'log all data (mainly for debug or perf tuning)
	Public SeqNum As Long = 0 'log entry sequence# (useful to check for missing/overwritten entries)
	Public StartTime As Date = Date.Now 'log entry times are relative to start of execution
	Public LogFile As String = String.Empty 'added for easier debug/performance analysis

	Private m_logbuf As New StringBuilder
	Private ReadOnly m_serial As New Object 'serialize writes to log file to avoid threading problems
	
'display error message and write to log:
	Public Sub ReportError(where As String, exc As Exception)
'NOTE to style critics: I hate line numbers but I also hate cluttering up source code with a lot of state/location info
'Line#s appear to be the most compact way to show the location of errors (since VB.NET doesn't have a __LINE__ macro like the C++ preprocessor), and there is no additional overhead
		If Erl <> 0 Then where &= "@" & Erl 'show line# if present
		Dim msg As String = String.Format("{0} (error {1} in {2})", Err.Description, Err.Number, where)
		If exc IsNot Nothing Then
			Me.LogMsg(String.Format("ERROR: {0}{1}{2}", msg, vbCRLF, exc.StackTrace), True)
			If exc.InnerException IsNot Nothing Then Me.LogMsg(String.Format("Inner exc: {0}{1}{2}", exc.InnerException.Message, vbCRLF, exc.InnerException.StackTrace), True)
		else
			Me.LogMsg(String.Format("ERROR: {0}", msg), True)
		End If
		If Not Me.ErrPause Then Return
'this isn't quite right yet, but at least it stops
'TODO: make all threads wait (to prevent a lot of windows during sequence execution)
		If MessageBox.Show(String.Format("{0}{1}Continue execution?", msg, vbCrLf), "ERROR", System.Windows.Forms.MessageBoxButtons.OKCancel, System.Windows.Forms.MessageBoxIcon.Question) <> DialogResult.OK Then Thread.CurrentThread.Abort
	End Sub

'dump a byte array as hex values:
	Public Function DumpHex(buf As Byte(), ActualLen As Integer, MaxLen As Integer) As String
		Dim hexbuf As New StringBuilder
		Dim trunc As Boolean = False, prevbyte As Byte = 0, prevcount As Integer = 0
'		While ActualLen > 1
'			If buf(ActualLen - 1) <> 0 Then Exit While
'			trunc = True
'			ActualLen -= 1
'		End While
		For i As Integer = 1 To ActualLen
'			If i > MaxLen Then trunc = True: Exit For
			If (i > 1) And (buf(i - 1) = prevbyte) Then prevcount += 1: If i < ActualLen Then Continue For
			If prevcount > 2 Then
				hexbuf.AppendFormat("*{0} ", prevcount)
			ElseIf prevcount = 2 Then 'skip the fancy notation
				hexbuf.AppendFormat("x{0:X2} ", prevbyte)
			End If
			If i < ActualLen Then hexbuf.AppendFormat("x{0:X2} ", buf(i-1))
			If hexbuf.Length > 4 * MaxLen Then hexbuf.Append(" ..."): Exit For
			prevbyte = buf(i - 1)
			prevcount = 1
		Next i
'		If trunc Then hexbuf.Append(" ...")
		Return hexbuf.ToString
	End Function

'write message to log file with relative timestamp:
	Public Sub LogMsg(buf As String)
		Me.LogMsg(buf, False)
	End Sub
	Public Sub LogMsg(buf As String, flush As Boolean)
		On Error Resume Next 'just ignore errors if they occur, rather than getting into an infinite loop
		Me.SeqNum += 1
		Dim elapsed As TimeSpan = DateTime.Now.Subtract(Me.StartTime)
		Dim msg as String = String.Format("{7}[{0} T+{1:D}:{2:D2}:{3:D2}.{4:D3} &{8:X}] {5}{6}", Me.SeqNum, elapsed.Hours, elapsed.Minutes, elapsed.Seconds, elapsed.Milliseconds, buf, vbCRLF, IIf(Me.SeqNum = 1, vbCRLF, String.Empty), Thread.CurrentThread.ManagedThreadId)
		SyncLock Me.m_serial 'avoid threading problems
			Me.m_logbuf.Append(msg)
			If (Me.m_logbuf.Length < &H800) And (Not flush) Then Return 'buffered for better performance
			If String.IsNullOrEmpty(Me.LogFile) Then 'nowhere to write it
				msg = Me.m_logbuf.ToString
			ElseIf Me.LogFile.Equals("console", StringComparison.CurrentCultureIgnoreCase) Then
				Console.Write(msg) 'msg already contains CRLF
				msg = String.Empty
			Else
'				Monitor.Enter(LogContention)
				System.IO.File.AppendAllText(Me.LogFile, Me.m_logbuf.ToString) 'AppendAllText is NOT THREAD-SAFE, so access must be managed
'				Monitor.Exit(LogContention)
				msg = String.Empty
			End If
			Me.m_logbuf.Length = 0 ' = New StringBuilder
		End SyncLock
'only show messages outside lock
		If Err.Number <> 0 Then MessageBox.Show(String.Format("{0} (error {1} while logging)", Err.Description, Err.Number), "LOG ERROR", System.Windows.Forms.MessageBoxButtons.OK, System.Windows.Forms.MessageBoxIcon.Exclamation)
		If Not String.IsNullOrEmpty(msg) Then MessageBox.Show(msg, "LOG MSG", System.Windows.Forms.MessageBoxButtons.OK, System.Windows.Forms.MessageBoxIcon.Information)
	End Sub
End Class
#End If


'helper functions:
Public Class FxUtils
'no string join method?
	Public Shared Function JoinString(delim As String, strary As String()) As String
		Dim sb As New StringBuilder
		For Each str As String In strary
			If String.IsNullOrEmpty(str) Then Continue For
			If sb.Length > 0 Then sb.Append(delim)
			sb.Append(str)
		Next str
		Return sb.ToString()
	End Function

'convert "# x #" string to size:
	Public Shared Function String2Size(whstr As String) As Size
'		Try
'		If String.IsNullOrEmpty(whstr) Then Return New Size(0, 0)
		Dim ofs As Integer = whstr.IndexOf("x")
		If ofs = - 1 Then Return New Size(0, 0)
		Return New Size(Convert.ToInt32(whstr.Substring(0, ofs).Trim()), Convert.ToInt32(whstr.Substring(ofs+1).Trim()))
'		Catch exc As Exception
'			Me.m_logger.ReportError(String.Format("invalid size string: {0}", whstr), exc)
'			Return New Size(0, 0)
'		End Try
	End Function
End Class

End Namespace
